/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#ifndef VXS_BOND_H
#define VXS_BOND_H

#include "Vec3D.h"
class CVXS_Voxel;
class CVX_Sim;

//!Bond types between two voxels.
/*!These define the bahavior of two voxels that interact in the simulation*/
enum BondType {
	B_LINEAR, //!< Generates forces and moments between connected voxels. Generates both compressive and tensile forces to resist both pushing together and pulling apart.
	B_LINEAR_CONTACT, //!< Generates repulsive forces when the voxels overlap. No tensile forces are generated. Voxels resist penetration but nothing holds them together.
	B_TILE_EDGE, //!< Reserved for future use.
	B_TILE_ZCORNER, //!< Reserved for future use.
	B_INPUT_LINEAR_NOROT //!<Generates forces but not moments between connected voxels. This is equivalent to a simple spring force between voxels.
};

class CVXS_Bond
{
public:
	CVXS_Bond(void);
public:
	~CVXS_Bond(void);

	CVXS_Bond& operator=(const CVXS_Bond& Bond); //overload "=" 
	CVXS_Bond(const CVXS_Bond& Bond) {*this = Bond;} //copy constructor

	void CalcForce(void); //calculates force, positive for tension, negative for compression
	
	//sub force calculation types...
	void CalcLinForce();
	void CalcContactForce();
	Vec3D CalcTileLookupForce(void);
	void CalcSimpleSpringForce();

	CVX_Sim* p_Sim;

	Vec3D Force1, Force2;
	Vec3D Moment1, Moment2;

	Vec3D RestDist; //From Voxel 1 to Voxel 2 (in global coordinate system... ie points in different directions for different bonds)
	Vec3D OrigDist; //original distance from Vox1 to Vox2 (in global CS). Does not change for non-linear stretched materials.

//	double Compression; //how much is this bond compressed? 1.0 is nominal, .5 is compressed 50% of original length, etc
//	double CurStress;
//	double CurGap;

//	Vec3D CurPVec; //current relative distance in position - in M, not mm
//	Vec3D CurTVec; //current relative difference in theta

	int Vox1SInd;
	int Vox2SInd;

	CVXS_Voxel* Vox1(void); //inline CVXS_Voxel* Vox1(void);
	CVXS_Voxel* Vox2(void); //inline CVXS_Voxel* Vox2(void);

	BondType ThisBondType;

//	void UpdateStatus(void);
//	void ResetStatus(void);

	bool Yielded; //has the bond yielded? (only update these after relaxation cycle!
	bool Broken; //has the bond broken?
	void SetYielded(void);
	void SetBroken(void);


//	double CurStiffness;


	//temp:
	//Vec3D Xh, Yh, Zh;

	//Constant parameters for this bond!
	bool CalcConsts(void); //fills in the constant parameters for the bond... returns false if unsensible material properties
	bool UpdateBondParams(vfloat CurStrainIn); //Updates yielded, brokem and CurStress based on CurStrainIn

	Vec3D L;
	vfloat E, u, G, A, Iy, Iz, J, a1, a2, b1y, b2y, b3y, b1z, b2z, b3z, CTE;
	CQuat ElRot;

	vfloat CurStrain, CurStress;
	vfloat MaxStrain; //the most strain this bond has undergone (needed for plastic deformation...)

	bool Perm; //is this bond permanent, (not temporary contact...)


	//these save about 15% time avoiding looking up pointers in the array!
	int _v1SIndCache, _v2SIndCache;
	CVXS_Voxel *_pV1Cache, *_pV2Cache;


};

//for nonlinear:
//http://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=3&ved=0CDkQFjAC&url=http%3A%2F%2Fsteel.sejong.ac.kr%2Fdown%2Fpaper%2Fi-paper-13.pdf&ei=fmFuUP_kAeOGyQGIrIDYDQ&usg=AFQjCNG3YZI0bd9OO69VQqV7PTO3KIsEyQ&cad=rja

//http://www.colorado.edu/engineering/cas/courses.d/IFEM.d/
//http://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=2&ved=0CCcQFjAB&url=http%3A%2F%2Fwww.eng.fsu.edu%2F~chandra%2Fcourses%2Feml4536%2FChapter4.ppt&ei=919uUKOnDamryQGc_oEI&usg=AFQjCNER15oW2k4KtNrNR3FdvxNnUbnRVw&cad=rja
	/*	1		2		3		4		5		6				7		8		9		10		11		12
		Kh[0][0] = a1;											Kh[0][6] = -a1;												| Px1
				Kh[1][1] = b1;					Kh[1][5] = b2;			Kh[1][7] = -b1;					Kh[1][11] = b2;		| Py1
						Kh[2][2] = b1;	Kh[2][4] = -b2;							Kh[2][8] = -b1;	Kh[2][10] = -b2;			| Pz1
								Kh[3][3] = a2;											Kh[3][9] = -a2;						| Ax1
						Kh[4][2] = -b2;	Kh[4][4] = 2*b3;						Kh[4][8] = b2;	Kh[4][10] = b3;				| Ay1
				Kh[5][1] = b2;					Kh[5][5] = 2*b3;		Kh[5][7] = -b2;					Kh[5][11] = b3;		| Az1
		Kh[6][0] = -a1;											Kh[6][6] = a1;												| Px2
				Kh[7][1] = -b1;					Kh[7][5] = -b2;			Kh[7][7] = b1;					Kh[7][11] = -b2;	| Py2
						Kh[8][2] = -b1;	Kh[8][4] = b2;							Kh[8][8] = b1;	Kh[8][10] = b2;				| Pz2
								Kh[9][3] = -a2;											Kh[9][9] = a2;						| Ax2
						Kh[10][2] = -b2;Kh[10][4] = b3;							Kh[10][8] = b2;	Kh[10][10] = 2*b3;			| Ay2
				Kh[11][1] = b2;					Kh[11][5] = b3;			Kh[11][7] = -b2;				Kh[11][11] = 2*b3;	| Az2

//http://www.eng-tips.com/viewthread.cfm?qid=14924&page=97
k[ 1][ 1] =  E*area/L;
k[ 1][ 7] = -k[1][1];
k[ 2][ 2] =  12.0*E*Iz/(L*L*L);
k[ 2][ 6] =  6.0*E*Iz/(L*L);
k[ 2][ 8] = -k[2][2];
k[ 2][12] =  k[2][6];
k[ 3][ 3] =  12.0*E*Iy/(L*L*L);
k[ 3][ 5] = -6.0*E*Iy/(L*L);
k[ 3][ 9] = -k[3][3];
k[ 3][11] =  k[3][5];
k[ 4][ 4] =  G*J/L;
k[ 4][10] = -k[4][4];
k[ 5][ 5] =  4.0*E*Iy/L;
k[ 5][ 9] =  6.0*E*Iy/(L*L);
k[ 5][11] =  2.0*E*Iy/L;
k[ 6][ 6] =  4.0*E*Iz/L;
k[ 6][ 8] = -6.0*E*Iz/(L*L);
k[ 6][12] =  2.0*E*Iz/L;
k[ 7][ 7] =  k[1][1];
k[ 8][ 8] =  k[2][2];
k[ 8][12] = -k[2][6];
k[ 9][ 9] =  k[3][3];
k[ 9][11] =  k[5][9];
k[10][10] =  k[4][4];
k[11][11] =  k[5][5];
k[12][12] =  k[6][6];
	
	*/
#endif //VXS_BOND_H
