/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#ifndef VXS_VOXEL_H
#define VXS_VOXEL_H

#include "Vec3D.h"
#include <vector>

//http://gafferongames.com/game-physics/physics-in-3d/

//Maximum number of bonds to a given voxel (including temporary contact bonds...)
#define NUMVBOND 50

class CVXS_Bond;
class CVX_Sim;
class CQuat;

struct VoxState {
	VoxState(void) {Pos = Vec3D(0,0,0); LinMom = Vec3D(0,0,0); Angle = CQuat(1.0, 0, 0, 0); AngMom = Vec3D(0,0,0); Scale = Vec3D(0,0,0); ScaleMom = Vec3D(0,0,0); Vel = Vec3D(0,0,0); AngVel = Vec3D(0,0,0); ScaleVel = Vec3D(0,0,0);};
	VoxState(const VoxState& SIn) {*this = SIn;} //copy constructor
	inline VoxState& operator=(const VoxState& SIn) {Pos = SIn.Pos; LinMom = SIn.LinMom; Angle = SIn.Angle; AngMom = SIn.AngMom; Scale = SIn.Scale; ScaleMom = SIn.ScaleMom; Vel = SIn.Vel; AngVel = SIn.AngVel; ScaleVel = SIn.ScaleVel; return *this; }; //overload equals

	Vec3D Pos; //translation
	Vec3D LinMom;

	CQuat Angle; //rotation
	Vec3D AngMom;
		
	Vec3D Scale; 
	Vec3D ScaleMom; //Scaling

	//convenience derived secondary quantittie:
	Vec3D Vel;
	Vec3D AngVel;
	Vec3D ScaleVel;

};

struct dVoxState { //derivative of voxel state
	dVoxState(void) {LinMom = Vec3D(0,0,0); Force = Vec3D(0,0,0); AngMom = Vec3D(0,0,0); Moment = Vec3D(0,0,0); ScaleMom = Vec3D(0,0,0); ScaleForce = Vec3D(0,0,0);};
	dVoxState(const dVoxState& dSIn) {*this = dSIn;} //copy constructor
	inline dVoxState& operator=(const dVoxState& dSIn) {LinMom = dSIn.LinMom; Force = dSIn.Force; AngMom = dSIn.AngMom; Moment = dSIn.Moment; ScaleMom = dSIn.ScaleMom; ScaleForce = dSIn.ScaleForce; return *this; }; //overload equals

	Vec3D LinMom;
	Vec3D Force;

	Vec3D AngMom;
	Vec3D Moment;
	
	Vec3D ScaleMom;
	Vec3D ScaleForce;
};


class CVXS_Voxel
{
public:
	CVXS_Voxel(void);
public:
	~CVXS_Voxel(void);

	CVXS_Voxel(const CVXS_Voxel& VIn) {*this = VIn;} //copy constructor
	CVXS_Voxel& operator=(const CVXS_Voxel& VIn);

protected:
	CVX_Sim* p_Sim;

public:
	VoxState S;
	VoxState PrevS; //needed for Verlet...
	dVoxState dS; //needed for RK4

	void LinkToVXSim(CVX_Sim* pSimIn) {p_Sim = pSimIn;} //link this voxel back to the simulator
	void UpdateDState(); //calculates the derivative of our current state "S". All bonds should be updates before this!
	void UpdateState(vfloat dt, dVoxState& Derivative); //advances this state "S" with previously calculated dS by dt

	void EulerStep();
	void VerletStep(vfloat dt);

	Vec3D CalcTotalForce(bool WithRestraint = true); //calculates InternalForce
	Vec3D CalcTotalMoment(void); //Calculates TotalMoment
	Vec3D CalcTotalScaleForce(void); //calculate Scaling force

	vfloat CurMaxStrain, CurMaxStress;
	
	//keep track of the bonds this voxel is participating in...
	std::vector<int> BondInds;
	CVXS_Bond* pBond(int LocalBondIndex);
	int NumLocalBonds(void){return (int)BondInds.size();};
	void UnlinkTmpBonds(void); //(assumes all perm at beginning...) removes any temprary bonds from our local list...

	std::vector<int> NearbyVoxInds;
	int NumNearbyVox(void){return (int)NearbyVoxInds.size();};
	CVXS_Voxel* pNearbyVox(int LocalNearbyInd);


	void CalcNearby(int NumHops); //populates how close of voxels to ignore in the lattice
	inline bool IsNearbyVox(int GlobalRVoxInd) {for (int i=0; i<NumNearbyVox(); i++) if (NearbyVoxInds[i] == GlobalRVoxInd) return true; return false;};
	inline bool IsMe(CVXS_Voxel* pCompare) {if (pCompare->MyXIndex == MyXIndex) return true; else return false;};

	//Static info about this voxel:
	bool Fixed; //is this voxel fixed (constrained to NomPos?)
	vfloat Mass; //The mass of this voxel
	vfloat Inertia; //rotational mass
	vfloat ScaleMass; //scaling mass
	int MyXIndex; //index in the original VXC Array
	int MySIndex; //index in the Relaxation voxel Array

	void SetMatIndex(int MatIndexIn) {MatIndex = MatIndexIn;}
	int GetMatIndex(void) {return MatIndex;}
private: //todo: all references thourgh access functions to channel blending info!
	int MatIndex; //Material associated with this voxel (pre blending_
public:
	void CalcMyBlendMix(); //requires p_sim, p_Sim->LocalVXC, and MyXIndex to already be in place...
	std::vector<vfloat> BlendMix; //array with size equal to number of materials in Palette. Contains percentage of each material. All numbers should add up to 1.
	void CalcVoxMatStress(vfloat StrainIn, vfloat* pStressOut, bool* IsPastYielded, bool* IsPastFail);
	vfloat GetEMod(void); //gets elastic modulus, taking blending into account
	vfloat GetPoisson(void); //gets poisson ratio, taking blending into account
	vfloat GetCTE(void); //gets CTE, taking blending into account


	Vec3D NomPos; //nominal position (Upon import. DO NOT change this, as it is needed for resetting/bond directions
	Vec3D NomScale; //nominal size (never changes)

	bool StaticFricFlag; //flag to set if this voxel shouldnot move in X/Y due to being in static friction regime


	Vec3D ExtForceApplied; //is there an external force applied to this voxel?

	//Other control infromation
//	bool Displacable; //is this voxel to be displaced in a SS test?
	Vec3D ExtDisplaceApplied; //displaced position, for prescribed strain
	bool VYielded;
	bool VBroken;
//	void ResetStatus(void);

	//Color stuff
	void SetGLColor(void);
	void SetColor(float r, float g, float b, float a);
	float m_Red; //so that different, say, BC's can have different colors... (not material colors...
	float m_Green;
	float m_Blue;
	float m_Trans;

	//Other effects that might have soemthing to do with force calculation
	Vec3D CalcFloorEffect(Vec3D TotalVoxForce); //calculates the object's interaction with a floor under gravity
	Vec3D CalcGNDDampEffect(void); //damps everything to ground as quick as possible...
	void CalcFluidEffect(void); //calculates the object's interaction if submersed in neutrally bouyant fluid
	vfloat FluidResistance; //the resistance (in the current direction of motion) due to fluid. Zero is no resistance (factor of 1.0), infinite is lack of motion (factor of 0.0)

	//for debuging
	Vec3D TotVoxForce;
	Vec3D TotVoxMoment;

};

#endif //VXS_VOXEL_H