/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#include "VX_Environment.h"

#ifdef USE_OPEN_GL
#ifdef QT_GUI_LIB
#include <qgl.h>
#else
#include "OpenGLInclude.h" //If not using QT's openGL system, make a header file "OpenGLInclude.h" that includes openGL library functions 
#endif
#include "GL_Utils.h"
#endif

CVX_Environment::CVX_Environment(void)
{
	pObj = NULL;

	GravEnabled = true;
	GravAcc = -9.81f; //m/s^2
	
	FloorEnabled = true;

	TempEnabled = true;
	VaryTempEnabled = false;
	TempAmp = 25;
	TempBase = 25; //degress celcius
	TempPeriod = 0.1f; //in seconds
	CurTemp = 25;
}

CVX_Environment::~CVX_Environment(void)
{
}

void CVX_Environment::SaveBCXFile(std::string filename)
{
	CXML_Rip XML;
	WriteXML(&XML);
	XML.SaveFile(filename);
}

bool CVX_Environment::LoadBCXFile(std::string filename)
{
	CXML_Rip XML;
	if (!XML.LoadFile(filename)) return false;
	ReadXML(&XML);
	return true;
}



void CVX_Environment::WriteXML(CXML_Rip* pXML)
{
	pXML->DownLevel("Environment");
		pXML->DownLevel("Fixed_Regions");
		pXML->Element("NumFixed", (int)Fixed.size());
		for (int i=0; i<(int)Fixed.size(); i++){
			Fixed[i].WriteXML(pXML);
		}
		pXML->UpLevel();

		pXML->DownLevel("Forced_Regions");
		pXML->Element("NumForced", (int)Forced.size());
		for (int i=0; i<(int)Forced.size(); i++){
			Forced[i].WriteXML(pXML);
		}
		pXML->UpLevel();

		pXML->DownLevel("Gravity");
		pXML->Element("GravEnabled", GravEnabled);
		pXML->Element("GravAcc", GravAcc);
		pXML->Element("FloorEnabled", FloorEnabled);

		pXML->UpLevel();

		pXML->DownLevel("Thermal");
		pXML->Element("TempEnabled", TempEnabled);
		pXML->Element("TempAmp", TempAmp);
		pXML->Element("TempBase", TempBase);
		pXML->Element("VaryTempEnabled", VaryTempEnabled);
		pXML->Element("TempPeriod", TempPeriod);
		pXML->UpLevel();
	pXML->UpLevel();
	
//	if(!OnlyBCs) pObj->WriteXML(pXML);
}

bool CVX_Environment::ReadXML(CXML_Rip* pXML, std::string* RetMessage) //pXML pointer to an "Environment" element
{
	ClearBCs();

	//Load fixed Regions
	if (pXML->FindElement("Fixed_Regions")){
		int NumFixedRegions;
		CVX_FRegion tmpRegion;

		if (!pXML->FindLoadElement("NumFixed", &NumFixedRegions)) NumFixedRegions = 0;
		for (int i=0; i<NumFixedRegions; i++){
			pXML->FindElement("FRegion");
			tmpRegion.ReadXML(pXML);
			Fixed.push_back(tmpRegion);
		}
		if (NumFixedRegions != 0) pXML->UpLevel(); //FRegion
		pXML->UpLevel(); //Fixed_Regions
	}
	
	//Load forced Regions
	if (pXML->FindElement("Forced_Regions")){
		int NumForcedRegions;
		CVX_FRegion tmpRegion;

		if (!pXML->FindLoadElement("NumForced", &NumForcedRegions)) NumForcedRegions = 0;
		for (int i=0; i<NumForcedRegions; i++){
			pXML->FindElement("FRegion");
			tmpRegion.ReadXML(pXML);
			Forced.push_back(tmpRegion);
		}
		if (NumForcedRegions != 0) pXML->UpLevel(); //FRegion
		pXML->UpLevel(); //Fixed_Regions
	}

	if (pXML->FindElement("Gravity")){
		if (!pXML->FindLoadElement("GravEnabled", &GravEnabled)) GravEnabled = false;
		if (!pXML->FindLoadElement("GravAcc", &GravAcc)) GravAcc = -9.81f;
		if (!pXML->FindLoadElement("FloorEnabled", &FloorEnabled)) FloorEnabled = false;

		pXML->UpLevel();
	}

	if (pXML->FindElement("Thermal")){
		if (!pXML->FindLoadElement("TempEnabled", &TempEnabled)) TempEnabled = false;
		if (!pXML->FindLoadElement("TempAmp", &TempAmp)) TempAmp = 25;
		if (!pXML->FindLoadElement("TempBase", &TempBase)) TempBase = 25;
		if (!pXML->FindLoadElement("VaryTempEnabled", &VaryTempEnabled)) VaryTempEnabled = false;
		if (!pXML->FindLoadElement("TempPeriod", &TempPeriod)) TempPeriod = 0.1f;
		CurTemp = TempBase;

		pXML->UpLevel();
	}



	return true;
}

/*! All voxels touching this region will be immobile in the simulation.
@param[in] Location The corner of the region closest to the origin. Specified as a percentage (in X, Y, and Z respectively) of the overall workspace. (Location.x, Location.y, and Location.z each have a range of [0.0 to 1.0]).
@param[in] Size The size of the region. Specified as a percentage (in X, Y, and Z respectively) of the overall workspace. (Size.x, Size.y, and Size.z each have a range of [0.0 to 1.0]).
*/
void CVX_Environment::AddFixedRegion(Vec3D Location, Vec3D Size) //fixes a region of voxels
{
	CVX_FRegion MyRegion;
	MyRegion.CreateBoxRegion(Location, Size);
	MyRegion.SetColor(0.0f, 0.0f, 0.0f, 1.0f);
	MyRegion.Fixed = true;
	Fixed.push_back(MyRegion);
}

/*! All voxels touching this region will have an external force applied to them. The provided Force vector will be divided equally between all voxels.
@param[in] Location The corner of the region closest to the origin. Specified as a percentage (in X, Y, and Z respectively) of the overall workspace. (Location.x, Location.y, and Location.z each have a range of [0.0 to 1.0]).
@param[in] Size The size of the region. Specified as a percentage (in X, Y, and Z respectively) of the overall workspace. (Size.x, Size.y, and Size.z each have a range of [0.0 to 1.0]).
@param[in] Force The force to be distributed across this region in Newtons. The force is divided equally among all voxels in the forced region.
*/
void CVX_Environment::AddForcedRegion(Vec3D Location, Vec3D Size, Vec3D Force) //applies a force to a region of voxels
{
	CVX_FRegion MyRegion;
	MyRegion.CreateBoxRegion(Location, Size);
	MyRegion.SetColor(0.0f, 0.0f, 0.0f, 1.0f);
	MyRegion.Force = Force;
	Forced.push_back(MyRegion);
}

void CVX_Environment::ClearBCs(void) //clears all fixed and forced regions
{
	Fixed.clear();
	Forced.clear();
}

void CVX_Environment::RemoveDisconnected(void) //removes regions not connected to ground... (and sets fixed list...) ALSO removes lone voxels (no bonds...)
{
	int NumFixed = 0;

	int NumVox = pObj->GetNumVox(); //number of voxels

	int* List = NULL; //list of voxels to expand
	List = new int[NumVox];
	int ListEnd = 0; //list iterators
	int ListBegin = 0;

	bool* Visited = NULL; //size of full voxel array, set to true if connected!
	Visited = new bool [pObj->GetStArraySize()];

	//Pupoluate fixed...
	int DOFInd = 0;
	Vec3D point;
	Vec3D size = pObj->GetLatDimEnv()/2.0;
	Vec3D WSSize = pObj->GetWorkSpace();

	for (int i=0; i<pObj->GetStArraySize(); i++){ //for all posible voxels:
		Visited[i] = false;

		if (pObj->GetMat(i) != 0){ //if there's a voxel here...
			point = pObj->GetXYZ(i);

			//Do fixed constraints
			for (int j = 0; j<(int)Fixed.size(); j++){ //go through each primitive defined as a constraint!
				if (!Visited[i] && Fixed[j].pRegion->IsTouching(&point, &size, &WSSize)){ //if this point is within
					NumFixed++;
					List[ListEnd++] = i; //add to our list to check later...
					Visited[i] = true;
				}
			}
			DOFInd++;
		}
		
	}

	//go through the list, adding neighbors to the list until we run out...
	int ThisIndex, ThatIndex;
	int x, y, z;
	bool FixedAlone;

	while (ListBegin != ListEnd){
		ThisIndex = List[ListBegin++];

		pObj->GetXYZNom(&x, &y, &z, ThisIndex);
		
		FixedAlone = true; //flag to see if this is a lone grounded one with no bonds... (true unless find a bond...)

		for (int i=0; i<6; i++){ //6 possible directions to find cube
			switch (i){ //go through each dimension
				case 0: ThatIndex = pObj->GetIndex(x+1, y, z); break; //+X direction
				case 1: ThatIndex = pObj->GetIndex(x-1, y, z); break; //-X direction
				case 2: ThatIndex = pObj->GetIndex(x, y+1, z); break; //+Y direction
				case 3: ThatIndex = pObj->GetIndex(x, y-1, z); break; //-Y direction
				case 4: ThatIndex = pObj->GetIndex(x, y, z+1); break; //+Z direction
				case 5: ThatIndex = pObj->GetIndex(x, y, z-1); break; //-Z direction
			}

			if (ThatIndex != -1 && pObj->GetMat(ThatIndex) != 0){ //if its there
				if (Visited[ThatIndex] == false){ //if we want to add it to the list
					Visited[ThatIndex] = true; //we've hit it now!
					List[ListEnd++] = ThatIndex;
				}
				FixedAlone = false; //we found a bond!
			}

			
		}

		if (FixedAlone) Visited[ThisIndex] = false; //flag for removal...
	}


	for (int i=0; i<pObj->GetStArraySize(); i++){ //for all posible voxels:
		if (Visited[i] == false) //if we never got there
			pObj->SetMat(i, 0); //erase non-connected material!
	}

//	DOF = pObj->GetNumVox()*DOFperBlock; //set this to our new DOF value

	//fill in fixed list... ((would be quicker to do this earlier when we checked everything against fixed regions
	//DOFInd = 0;
	//for (int i=0; i<pObj->GetStArraySize(); i++){ //for all posible voxels:
	//	if (pObj->GetMat(i) != 0){ //if there's a voxel here...
	//		FixedList[DOFInd] = false; //assume not fixed
	//		point = pObj->GetXYZPerc(i);

	//		//Do fixed constraints
	//		for (int j = 0; j<(int)Fixed.size(); j++){ //go through each primitive defined as a constraint!
	//			if (Fixed[j].pRegion->IsTouching(&point, &size)){ //if this point is within
	//				FixedList[DOFInd] = true; //set this one as fixed
	//			}
	//		}
	//		DOFInd++;
	//	}
	//}

	delete[] List;
	List = NULL;
	delete[] Visited;
	Visited = NULL;
}
//pass in the global time, and a pointer to the local object so its material temps can be modified
void CVX_Environment::UpdateCurTemp(vfloat time, CVX_Object* LocalVXC)
{
	if (VaryTempEnabled){
		CurTemp = TempBase + (TempAmp-TempBase)*sin(2*3.1415926f/TempPeriod*time);	//update the global temperature
		//now update the individual temperatures of each material (they can each have a different temperature)
		int NumObjMatrls = LocalVXC->GetNumMaterials();	//get the number of materials in the object
		for (int i = 0; i<NumObjMatrls; i++){
			LocalVXC->GetBaseMat(i)->SetMatTemp(TempBase + (TempAmp-TempBase)*sin((2*3.1415926f/TempPeriod) * time + LocalVXC->GetBaseMat(i)->GetMatTempPhase()));	//and update each one
			}
	}
	else {
		CurTemp = TempAmp;
		int NumObjMatrls = LocalVXC->GetNumMaterials();	//get the number of materials in the object
		for (int i = 0; i<NumObjMatrls; i++){
			LocalVXC->GetBaseMat(i)->SetMatTemp(TempAmp);	//and update each one
			}
		}
}


#ifdef USE_OPEN_GL
void CVX_Environment::DrawBCs(int Selected) //draws the current boundary conditions
{
	Vec3D WS = pObj->GetWorkSpace();
	for (int j=0; j<(int)Fixed.size(); j++){ //draw fixed constraints
		int MyGLindex = BCFIXED_GLIND_OFF+j;
		if (Selected == MyGLindex) Fixed[j].SetColor(0.0, 1.0, 0.0, 1.0); //selected
		else Fixed[j].SetColor(0.4, 0.6, 0.4, 1.0); //unselected

		glLoadName (MyGLindex); //to enable picking
		CPrimitive& Reg = *Fixed[j].pRegion;
		Fixed[j].DrawScaled(&WS);

		//todo:
		if (Fixed[j].Displace.Length2() != 0){
			Vec3D NForce = Fixed[j].Displace/Fixed[j].Displace.Length()*pObj->GetLatticeDim()*1.5; //normalize to 1.5 vox length
			for (int x=0; x<2; x++){ for (int y=0; y<2; y++){ for (int z=0; z<2; z++){
				Vec3D CurCorner = Vec3D((Reg.X+x*Reg.dX)*WS.x, (Reg.Y+y*Reg.dY)*WS.y, (Reg.Z+z*Reg.dZ)*WS.z);
				CGL_Utils::DrawArrow(CurCorner, NForce, CColor(0.5, 0.5, 0.5, 1.0));
			}}}
		}
	}
	for (int j=0; j<(int)Forced.size(); j++){ //draw force constraints
		int MyGLindex = BCFORCE_GLIND_OFF+j;
		if (Selected == MyGLindex) Forced[j].SetColor(1.0, 0.0, 1.0, 1.0); //selected
		else Forced[j].SetColor(0.6, 0.4, 0.6, 1.0); //unselected

		glLoadName (MyGLindex); //to enable picking
		CPrimitive& Reg = *Forced[j].pRegion;
		Forced[j].DrawScaled(&WS);

		//todo:
		Vec3D NForce = Forced[j].Force/Forced[j].Force.Length()*pObj->GetLatticeDim()*1.5; //normalize to 1.5 vox length
		for (int x=0; x<2; x++){ for (int y=0; y<2; y++){ for (int z=0; z<2; z++){
			Vec3D CurCorner = Vec3D((Reg.X+x*Reg.dX)*WS.x, (Reg.Y+y*Reg.dY)*WS.y, (Reg.Z+z*Reg.dZ)*WS.z);
			CGL_Utils::DrawArrow(CurCorner, NForce, CColor(0.5, 0.5, 0.5, 1.0));
		}}}
	}
}

#endif