/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#ifndef CVX_ENVIRONMENT_H
#define CVX_ENVIRONMENT_H

#include "VX_FRegion.h"
#include "VX_Object.h"

#define BCFORCE_GLIND_OFF 100000000
#define BCFIXED_GLIND_OFF 100001000

//!Describes physical environment of voxel object	
/*!Contains information about the boundary conditions (fixed and forced voxels) as well as parameters fro gravity, temperature, etc.*/
class CVX_Environment
{
public:
	CVX_Environment(void);
	~CVX_Environment(void);

	void AddObject(CVX_Object* pObjIn) {pObj = pObjIn;} //!< Links a voxel object to this environment. Only one voxel object may be linked at a time. @param[in] pObjIn Pointer to an initialized voxel object to link to this simulation.
	CVX_Object* pObj; //link to the object in this environment

	void SaveBCXFile(std::string filename);
	bool LoadBCXFile(std::string filename);

	//writes just local information
	void WriteXML(CXML_Rip* pXML);
	bool ReadXML(CXML_Rip* pXML, std::string* RetMessage = NULL);

	//boundary conditions
	std::vector <CVX_FRegion> Fixed;
	std::vector <CVX_FRegion> Forced;

	void AddFixedRegion(Vec3D Location, Vec3D Size); //!< Adds a region of voxels to be fixed to ground.
	void AddFixedRegion(CVX_FRegion* pRegion){CVX_FRegion tmp = *pRegion; tmp.Fixed = true; Fixed.push_back(tmp);} //!< Adds a pre-exisiting region as fixed. @param[in] pRegion Pointer to an initialized regions to be added as a fixed region.
	void DelFixedRegion(int Index){if(Index>=0 && Index<(int)Fixed.size()) Fixed.erase(Fixed.begin()+Index);} //!< Deletes a fixed region at the specified index in the Fixed region array. @param[in] Index Index of Fixed array to delete.
	void AddForcedRegion(Vec3D Location, Vec3D Size, Vec3D Force); //!< Applies a force to a region of voxels.
	void AddForcedRegion(CVX_FRegion* pRegion){CVX_FRegion tmp = *pRegion; tmp.Fixed = false; Forced.push_back(tmp);} //!< Adds a pre-exisiting region as forced with zero force. @param[in] pRegion Pointer to an initialized regions to be added as a forced region.
	void DelForcedRegion(int Index){if(Index>=0 && Index<(int)Forced.size()) Forced.erase(Forced.begin()+Index);} //!< Deletes a forced region at the specified index in the Forced region array. @param[in] Index Index of Forced array to delete.
	void ClearBCs(void); //!< Clears all fixed and forced regions
	int GetNumFixed(void) {return (int)Fixed.size();} //!< Returns the number of fixed regions currently in the environment.
	CVX_FRegion* GetFixedRegion(int index) {if(index>=0 && index<(int)Fixed.size()) return &Fixed[index]; else return NULL;} //1< Returns a temporary pointer to the fixed region at the specified index. @param[in] index Index of Fixed array element to return a pointer to.
	int GetNumForced(void) {return (int)Forced.size();} //!< Returns the number of forced regions currently in the environment.
	CVX_FRegion* GetForcedRegion(int index) {if(index>=0 && index<(int)Forced.size()) return &Forced[index]; else return NULL;}//!< Returns a temporary pointer to the forced  region at the specified index. @param[in] index Index of Forced array element to return a pointer to.

	void RemoveDisconnected(void); //!< Removes all voxels in the voxel object not connected to a fixed voxel. This is necessary to create non-singular matrices for static FEA analysis.

	//Other environment vars:
	bool GravEnabled;
	vfloat GravAcc; //m/s^2

	bool FloorEnabled; //do we want a floor system?
	
	bool TempEnabled, VaryTempEnabled;
	vfloat TempAmp, TempBase, TempPeriod; //degress celcius
	vfloat CurTemp; //updated based on time...
	vfloat CurTime;	//just a holder for the global time that is passed in

	vfloat GetCurTemp() {return CurTemp;};
	void UpdateCurTemp(vfloat time, CVX_Object* LocalVXC);
	void UpdateCurTime(vfloat time) {CurTime=time;};

#ifdef USE_OPEN_GL
	void DrawBCs(int Selected); //draws the current boundary conditions
#endif
};

#endif //CVX_ENVIRONMENT_H