/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#include "VX_Sim.h"
#include "VXS_Voxel.h"
#include "VXS_Bond.h"

#include <sstream>

#ifdef USE_OPEN_GL
#include "GL_Utils.h"
#endif


CVX_Sim::CVX_Sim(void)
{
	pEnv = NULL;

	print_scrn=false;
	WriteFitnessFile=false;
	FitnessRepType=0;	//no reporting is default

	CurColSystem = COL_SURFACE_HORIZON;
	CurIntegrator = I_EULER;

	PoissonKickBackEnabled = false;
	EnforceLatticeEnabled = false;

	SelfColEnabled = false;
	CollisionHorizon = 3.0;

	MaxVelLimitEnabled = false;
	MaxVoxVelLimit = (vfloat)0.1;

	BlendingEnabled = false;
	MixRadius=0.0;
	BlendModel=MB_LINEAR;
	PolyExp = 1.0;

	FluidDampEnabled = false;

	PlasticityEnabled = true;
	FailureEnabled = true;

	InputVoxSInd = -1;
	InputBondInd = -1;

	BondDampingZ = 1;
	ColDampingZ = 1;
	SlowDampingZ = (vfloat)0.001;
//	ViscousDamp = 0.0001;

	DtFrac = (vfloat)0.9; //percent of maximum dt to use


#ifdef USE_OPEN_GL
	ViewGeoMesh = true;
	ViewSurfMesh = false;
	ViewGeo = false;
	ViewBonds = false;
	ViewForce = false;
	ViewAngles = false;
	CurViewCol = RVC_MATERIAL;
#endif

	ClearAll();

}

CVX_Sim::~CVX_Sim(void)
{
	ClearAll();

//	out.close();
}

CVX_Sim& CVX_Sim::operator=(const CVX_Sim& rSim) //overload "=" 
{
	//TODO: set everything sensible equal.

	return *this;
}

void CVX_Sim::SaveVXAFile(std::string filename)
{
	CXML_Rip XML;
	WriteVXA(&XML);
	XML.SaveFile(filename);
}

void CVX_Sim::SaveResultFile(std::string filename)
{
	CXML_Rip XML;
	WriteResultFile(&XML);
	XML.SaveFile(filename);
}

bool CVX_Sim::LoadVXAFile(std::string filename)
{
	CXML_Rip XML;
	if (!XML.LoadFile(filename)) return false;
	ReadVXA(&XML);
	return true;
}

void CVX_Sim::WriteVXA(CXML_Rip* pXML)
{
	pXML->DownLevel("VXA");
	pXML->SetElAttribute("Version", "1.0");
	WriteXML(pXML);
	pEnv->WriteXML(pXML);
	pEnv->pObj->WriteXML(pXML);
	pXML->UpLevel();
}

void CVX_Sim::WriteResultFile(CXML_Rip* pXML)
{
	pXML->DownLevel("Voxelize_Sim_Result");
	pXML->SetElAttribute("Version", "1.0");
	pXML->DownLevel("Fitness");
	pXML->Element("FinalCOM_Dist", (GetCM()-IniCM).Length());
	pXML->UpLevel();
	pXML->UpLevel();

}

bool CVX_Sim::ReadVXA(CXML_Rip* pXML, std::string* RetMessage) //pointer to VXA element
{
//	pObj->ClearMatter();
	std::string ThisVersion = "1.0";
	std::string Version;
	pXML->GetElAttribute("Version", &Version);
	if (atof(Version.c_str()) > atof(ThisVersion.c_str())) if (RetMessage) *RetMessage += "Attempting to open newer version of VXA file!\n";

	if (pXML->FindElement("Simulator")){
		ReadXML(pXML);
		pXML->UpLevel();
	}

	//load environment
	if (pEnv && pXML->FindElement("Environment")){
		pEnv->ReadXML(pXML);
		pXML->UpLevel();
	}

	//Load VXC if pObj is valid...
	if (pEnv->pObj && (pXML->FindElement("VXC") || pXML->FindElement("DMF"))){
		pEnv->pObj->ReadXML(pXML);
		pXML->UpLevel();
	}
	return true;
}

void CVX_Sim::WriteXML(CXML_Rip* pXML)
{
	pXML->DownLevel("Simulator");
		pXML->DownLevel("Integration");
		pXML->Element("Integrator", (int)CurIntegrator);
		pXML->Element("DtFrac", DtFrac);
		pXML->UpLevel();

		pXML->DownLevel("Damping");
		pXML->Element("BondDampingZ", BondDampingZ);
		pXML->Element("ColDampingZ", ColDampingZ);
		pXML->Element("SlowDampingZ", SlowDampingZ);
		pXML->UpLevel();

		pXML->DownLevel("Collisions");
		pXML->Element("SelfColEnabled", SelfColEnabled);
		pXML->Element("ColSystem", CurColSystem);
		pXML->Element("CollisionHorizon", CollisionHorizon);
		pXML->UpLevel();

		pXML->DownLevel("Features");
		pXML->Element("MaxVelLimitEnabled", MaxVelLimitEnabled);
		pXML->Element("MaxVoxVelLimit", MaxVoxVelLimit);
		pXML->Element("BlendingEnabled", BlendingEnabled);
		pXML->Element("MixRadius", MixRadius);
		pXML->Element("BlendModel", BlendModel);
		pXML->Element("PolyExp", PolyExp);
		pXML->Element("FluidDampEnabled", FluidDampEnabled);
		pXML->Element("PoissonKickBackEnabled", PoissonKickBackEnabled);
		pXML->Element("EnforceLatticeEnabled", EnforceLatticeEnabled);
		pXML->UpLevel();


		pXML->DownLevel("SurfMesh");
		SurfMesh.DefMesh.WriteXML(pXML, true);
		pXML->UpLevel();

	pXML->UpLevel();

}

bool CVX_Sim::ReadXML(CXML_Rip* pXML, std::string* RetMessage)
{
	int tmpInt;

	if (pXML->FindElement("Integration")){
		if (pXML->FindLoadElement("Integrator", &tmpInt)) CurIntegrator = (IntegrationType)tmpInt; else CurIntegrator = I_EULER;
		if (!pXML->FindLoadElement("DtFrac", &DtFrac)) DtFrac = (vfloat)0.9;
		pXML->UpLevel();
	}
		
	if (pXML->FindElement("Damping")){
		if (!pXML->FindLoadElement("BondDampingZ", &BondDampingZ)) BondDampingZ = 1.0;
		if (!pXML->FindLoadElement("ColDampingZ", &ColDampingZ)) ColDampingZ = 1.0;
		if (!pXML->FindLoadElement("SlowDampingZ", &SlowDampingZ)) SlowDampingZ = 1.0;
		pXML->UpLevel();
	}

	if (pXML->FindElement("Collisions")){
		if (!pXML->FindLoadElement("SelfColEnabled", &SelfColEnabled)) SelfColEnabled = false;
		if (pXML->FindLoadElement("ColSystem", &tmpInt)) CurColSystem = (ColSystem)tmpInt; else CurColSystem = COL_SURFACE_HORIZON;
		if (!pXML->FindLoadElement("CollisionHorizon", &CollisionHorizon)) CollisionHorizon = (vfloat)2.0;
		pXML->UpLevel();
	}

	if (pXML->FindElement("Features")){
		if (!pXML->FindLoadElement("MaxVelLimitEnabled", &MaxVelLimitEnabled)) MaxVelLimitEnabled = false;
		if (!pXML->FindLoadElement("MaxVoxVelLimit", &MaxVoxVelLimit)) MaxVoxVelLimit = (vfloat)0.1;
		if (!pXML->FindLoadElement("BlendingEnabled", &BlendingEnabled)) BlendingEnabled = false;
		if (!pXML->FindLoadElement("MixRadius", &MixRadius)) MixRadius = 0.0;
		if (pXML->FindLoadElement("BlendModel", &tmpInt)) BlendModel = (MatBlendModel)tmpInt; else BlendModel = MB_LINEAR;
		if (!pXML->FindLoadElement("PolyExp", &PolyExp)) PolyExp = 1.0;

		if (!pXML->FindLoadElement("FluidDampEnabled", &FluidDampEnabled)) FluidDampEnabled = false;
		if (!pXML->FindLoadElement("PoissonKickBackEnabled", &PoissonKickBackEnabled)) PoissonKickBackEnabled = false;
		if (!pXML->FindLoadElement("EnforceLatticeEnabled", &EnforceLatticeEnabled)) EnforceLatticeEnabled = false;
		pXML->UpLevel();
	}

	if (pXML->FindElement("SurfMesh")){
		if (pXML->FindElement("CMesh")){
			SurfMesh.DefMesh.ReadXML(pXML);
			pXML->UpLevel();
		}
		pXML->UpLevel();
	}

	if (pXML->FindElement("SimStop")){
		if (!pXML->FindLoadElement("SimStopTime", &SimStopTime)) SimStopTime = FLT_MAX;
		if (!pXML->FindLoadElement("SimStopStep", &SimStopStep)) SimStopStep = LONG_MAX;
		pXML->UpLevel();
		}
	else {
		SimStopTime = FLT_MAX;
		SimStopStep = 2;	//go through only two steps, since no simulation was specified 
		}

	if (pXML->FindElement("Fitness")){
		int dummy_sdf;	//ignore the return, we're just looking for tag presence
		if (pXML->FindLoadElement("TrackCOM", &dummy_sdf)) FitnessRepType = 1;
		if (pXML->FindLoadElement("TrackVoxel", &TrackVoxel)) FitnessRepType = 2;
		if (pXML->FindLoadElement("FitnessFileNm", &FitnessFileNm)){ 
			WriteFitnessFile = true;
			}

		pXML->UpLevel();
		}


	return true;
}


void CVX_Sim::ClearAll(void) //Reset all initialized variables
{
	Initalized = false;
	LocalVXC.ClearMatter();

	//This should be all the stuff set by "Import()"
	VoxArray.clear();
	BondArray.clear();
	XtoSIndexMap.clear();
	StoXIndexMap.clear();
	SurfVoxels.clear();

	OrigVoxArray.clear();
	A.clear();
	B.clear();
	C.clear();
	D.clear();
	OrigBondArray.clear();

	MaxDispSinceLastBondUpdate = (vfloat)9e9; //arbitrarily high as a flag to populate bonds

	dt = 0.0; //calculated per-step

	TotalObjForce = Vec3D (0,0,0);
	TotalObjDisp = Vec3D (0,0,0);
	NormObjForce = 0;
	NormObjDisp = 0;
	MaxVoxDisp = 0.0f;
	MaxVoxVel = 0.0f;
	MaxBondStrain = 0.0f;
	MaxBondStress = 0.0f;

	IniCM = Vec3D(0,0,0);
	CurCM = Vec3D(0,0,0);

	MaxVoxVel = 0.0f;
//	BlendingEnabled = false;
//	MixRadius = 0.0;
//	BlendModel = MB_LINEAR;
//	PolyExp = 1.0;

	CurXSel = -1;
	Dragging = false;

//	SurfMesh.Clear();
//	VoxMesh.Clear();
}

int CVX_Sim::NumVox(void) 
{
	return (int)VoxArray.size()-1; //-1 is because we always have an input voxel shadow...
}

int CVX_Sim::NumBond(void) 
{
	return (int)BondArray.size();
}

void CVX_Sim::DeleteTmpBonds(void)
{
	int iT = NumVox();
	for (int i=0; i<iT; i++){
		VoxArray[i].UnlinkTmpBonds();
	}
	int NumBonds = NumBond();
	while (!BondArray[NumBonds-1].Perm){
		BondArray.pop_back();
		NumBonds--;
	}
}

CVXS_Voxel* CVX_Sim::InputVoxel(void)
{
	return &VoxArray[InputVoxSInd];
}

CVXS_Bond* CVX_Sim::InputBond(void)
{
	return &BondArray[InputBondInd];
}

/*! The environment should have been previously initialized and linked with a single voxel object. 
This function sets or resets the entire simulation with the new environment.
@param[in] pEnvIn A pointer to initialized CVX_Environment to import into the simulator.
@param[out] RetMessage A pointer to initialized string. Output information from the Import function is appended to this string.
*/
void CVX_Sim::Import(CVX_Environment* pEnvIn, CMesh* pSurfMeshIn, std::string* RetMessage)
{
	ClearAll(); //clears out all arrays and stuff

	if (pEnvIn != NULL) pEnv = pEnvIn;
	if (pEnv == NULL) {if (RetMessage) *RetMessage += "Invalid Environment pointer"; return;}

	LocalVXC = *pEnv->pObj; //make a copy of the reference digital object!

	int Index = 0; //keep track of how many voxel we've added (for storing reverse lookup array...)

	//initialize XtoSIndexMap & StoXIndexMap
	XtoSIndexMap.resize(LocalVXC.GetStArraySize(), -1); // = new int[LocalVXC.GetStArraySize()];
	StoXIndexMap.resize(LocalVXC.GetNumVox(), -1); // = new int [m_NumVox];

	Vec3D BCpoint;
	Vec3D BCsize = pEnv->pObj->GetLatDimEnv()/2.0;
	Vec3D WSSize = pEnv->pObj->GetWorkSpace();

	//count the number of voxels in each force region
	int* Sizes = new int[(int)pEnv->Forced.size()]; //holds how many voxels are within each forcing region
	for (int i=0; i<(int)pEnv->Forced.size(); i++) Sizes[i] = 0;
	for (int i=0; i<pEnv->pObj->GetStArraySize(); i++){
		if (pEnv->pObj->GetMat(i) != 0){ //if there's a voxel here...
			BCpoint = pEnv->pObj->GetXYZ(i);
			for (int j=0; j<(int)pEnv->Forced.size(); j++){ //for each forcing region
				if (pEnv->Forced[j].pRegion->IsTouching(&BCpoint, &BCsize, &WSSize)){
					Sizes[j]++;
				}
			}
		}
	}

	//Build voxel list
	for (int i=0; i<LocalVXC.GetStArraySize(); i++){ //for each voxel in the array
		XtoSIndexMap[i] = -1; //assume there is not a voxel here...

		if(LocalVXC.Structure[i] != 0 ){ //if there's material here
			CVXS_Voxel CurVox;
			CurVox.LinkToVXSim(this);

			CurVox.SetMatIndex(abs(LocalVXC.GetLeafMatIndex(i))); //go directly to leaf material index
			LocalVXC.GetXYZ(&CurVox.S.Pos, i);//Get XYZ location

			XtoSIndexMap[i] = Index; //so we can find this voxel based on it's original index
			StoXIndexMap[Index] = i; //so we can find the original index based on its simulator position

			//evaluate for removal:
			CurVox.NomPos = CurVox.S.Pos; //keep in case fixed...
			CurVox.NomScale = LocalVXC.GetLatDimEnv();
			CurVox.S.Scale = CurVox.NomScale;

			//calculate now to avoid doing it all later...
			CurVox.Mass = pow(LocalVXC.GetLatticeDim(), 3) * LocalVXC.GetLeafMat(i)->GetDensity(); //render density to mass here so we don't have to keep doing it.
			CurVox.Inertia = CurVox.Mass * pow(LocalVXC.GetLatticeDim(), 2)/6;
			CurVox.ScaleMass = CurVox.Mass; // * pow(LocalVXC.GetLatticeDim(), 4)*6/16;

			BCpoint = pEnv->pObj->GetXYZ(i);
			//Fix it if in fixed region
			for (int j = 0; j<(int)pEnv->Fixed.size(); j++){ //go through each primitive defined as a constraint!
				if (pEnv->Fixed[j].pRegion->IsTouching(&BCpoint, &BCsize, &WSSize)){ //if this point is within
					CurVox.Fixed = true;
					CurVox.ExtDisplaceApplied = pEnv->Fixed[j].Displace;
				}
			}

			//Force it if in forced region
			for (int j=0; j<(int)pEnv->Forced.size(); j++){
				if (pEnv->Forced[j].pRegion->IsTouching(&BCpoint, &BCsize, &WSSize)){
					CurVox.ExtForceApplied = pEnv->Forced[j].Force/Sizes[j];
				}
			}

			CurVox.MyXIndex = i;
			CurVox.MySIndex = Index;

			CurVox.PrevS = CurVox.S;

			if(BlendingEnabled) CurVox.CalcMyBlendMix(); //needs to be done basically last.

			VoxArray.push_back(CurVox);

			Index++;
		}
	}

	delete[] Sizes;
	Sizes = NULL;

	//add input voxel so that NumVox() works!
	InputVoxSInd = (int)VoxArray.size();
	CVXS_Voxel TmpVox;
	TmpVox.LinkToVXSim(this);
	VoxArray.push_back(TmpVox);


	//set up bonds...
	Vec3D RelDist;

	//exhaustive (slower, but still OK)
	for (int i=0; i<NumVox(); i++){ //for each voxel in our newly-made array
		for (int j = i+1; j<NumVox(); j++){
			if (LocalVXC.Voxel.GetVoxName() == VS_SPHERE || LocalVXC.Voxel.GetVoxName() == VS_BOX){ //HACK for box. re-do for non-cubic boxes!
				if (LocalVXC.IsAdjacent(StoXIndexMap[i], StoXIndexMap[j], false, &RelDist)){
				
//					if (!CreateBond(B_LINEAR, i, j, true) && RetMessage) *RetMessage += "At least one bond creation failed during import";


					//TMP!!!
					if (RelDist.x > 0.00001){
						if (!CreateBond(B_LINEAR, i, j, true) && RetMessage) *RetMessage += "At least one bond creation failed during import";
					}
					else {
						if (!CreateBond(B_LINEAR, j, i, true) && RetMessage) *RetMessage += "At least one bond creation failed during import";
					}	///
				}
			}
			else if (LocalVXC.Voxel.GetVoxName() == VS_VOXFILE){ // && LocalVXC.Voxel.File == "" //if special voxel shapes...
				////in plane bond-type
				//RelDist = CVoxelArray[j].Pos - CVoxelArray[i].Pos; //relative distance between voxels
				//if (abs(RelDist.z)<LocalVXC.Lattice.Lattice_Dim*LocalVXC.Lattice.Z_Dim_Adj/2 && (RelDist.ProjXY()).Length() < LocalVXC.Lattice.Lattice_Dim*1.2){ //if in the same plane
				//	CreateBond(B_TILE_EDGE, i, j, true); //add bond!
				//}
				////out-of-plane bond type
				//else if (abs(RelDist.z)<LocalVXC.Lattice.Lattice_Dim*LocalVXC.Lattice.Z_Dim_Adj*1.2 && (RelDist.ProjXY()).Length() < LocalVXC.Lattice.Lattice_Dim*1.2){
				//	CreateBond(B_TILE_ZCORNER, i, j, true); //add bond!
				//}

			}
			else {
				if (RetMessage) *RetMessage += "Unsupported voxel type.";
				return;
			}
		}
	}

	//Set up our surface list...
	for (int i=0; i<NumVox(); i++){ //for each voxel in our newly-made array
		if (VoxArray[i].NumLocalBonds() != 6){
			SurfVoxels.push_back(i);
		}

		//todo: only do for those on surfacem, I think.
		VoxArray[i].CalcNearby((int)(CollisionHorizon*1.5)); //populate the nearby array
	}

	//add input bond:
	InputBondInd = NumBond();
	CVXS_Bond TmpBond;
	TmpBond.Perm = true;
	TmpBond.p_Sim = this;
	TmpBond.ThisBondType = B_INPUT_LINEAR_NOROT;
	TmpBond.Vox1SInd = InputVoxSInd;
	TmpBond.Vox2SInd = InputVoxSInd;
	BondArray.push_back(TmpBond);

#ifdef USE_OPEN_GL
	VoxMesh.ImportLinkSim(this);
	VoxMesh.DefMesh.DrawSmooth = false;

//	(if pSurfMeshIn is null, use what we already have...)
//	pSurfMeshIn link up here...
	//if the input mesh is not valid, use marching cubes to create one
	CMesh RobMesh;
	if (!this->SurfMesh.DefMesh.Exists()){
		CArray3Df OccupancyArray(this->pEnv->pObj->GetVXDim(), this->pEnv->pObj->GetVYDim(), this->pEnv->pObj->GetVZDim()); 
		for (int g=0; g<OccupancyArray.GetFullSize(); g++){
			OccupancyArray[g]=this->pEnv->pObj->Structure.GetData(g);
			}
		CMarchCube::SingleMaterial(&RobMesh, &OccupancyArray, 0.05, this->pEnv->pObj->GetLatticeDim());
		SurfMesh.ImportSimWithMesh(this, &RobMesh);
		}
	else SurfMesh.ImportSimWithMesh(this, pSurfMeshIn);

//	if (pSurfMeshIn) 
//	else 
	

#endif

	ResetSimulation();

	Initalized = true;
	std::string tmpString;

	std::ostringstream os;
	os << "Completed Simulation Import: " << NumVox() << " Voxels, " << NumBond() << "Bonds.\n";
	*RetMessage += os.str();

}

/*! This bond is appended to the master bond array (BondArray). 
The behavior of the bond is determined by BondType. If the bond is permanent and should persist throughout the simulation PermIn should be to true.
@param[in] BondTypeIn The physical behavior of the bond being added.
@param[in] SIndex1 One simulation voxel index to be joined.
@param[in] SIndex2 The other simulation voxel index to be joined.
@param[in] PermIn Denotes whether this bond should persist throughout the simulation (true) or is temporary (false).
*/
bool CVX_Sim::CreateBond(BondType BondTypeIn, int SIndex1, int SIndex2, bool PermIn ) //take care of all dynamic memory stuff...
{
	CVXS_Bond tmp;

	tmp.p_Sim = this;

	tmp.ThisBondType = BondTypeIn;
	tmp.Vox1SInd = SIndex1;
	tmp.Vox2SInd = SIndex2;
	tmp.OrigDist = VoxArray[SIndex2].S.Pos - VoxArray[SIndex1].S.Pos;
//	tmp.RestDist = tmp.OrigDist;
	tmp.Perm = PermIn; 

	VoxArray[SIndex1].BondInds.push_back(NumBond());
	VoxArray[SIndex2].BondInds.push_back(NumBond());

	tmp.CalcConsts(); //do this here, cause we can...

	BondArray.push_back(tmp);

	return true;
}

void CVX_Sim::ResetSimulation(void)
{
	int iT = NumVox();
	for (int i=0; i<iT; i++){
		VoxArray[i].S = VoxState();
		VoxArray[i].dS = dVoxState();

		VoxArray[i].S.Pos = VoxArray[i].NomPos;
		VoxArray[i].S.Scale = VoxArray[i].NomScale;
		
		VoxArray[i].PrevS = VoxArray[i].S;

		VoxArray[i].VBroken = false;
		VoxArray[i].VYielded = false;
	}
	iT = NumBond();
	for (int j=0; j<iT; j++){
		BondArray[j].Yielded = false;
		BondArray[j].Broken = false;
		BondArray[j].RestDist = BondArray[j].OrigDist;

		BondArray[j].Force1 = Vec3D(0,0,0);
		BondArray[j].Force2 = Vec3D(0,0,0);
		BondArray[j].Moment1 = Vec3D(0,0,0);
		BondArray[j].Moment2 = Vec3D(0,0,0);
	}
}

/*! Given the current state of the simulation (Voxel positions and velocities) and information about the current environment, advances the simulation by the maximum stable timestep. 
The integration scheme denoted by the CurIntegrator member variable is used.
Calculates some relevant system statistics such as maximum displacements and velocities and total force.
Returns true if the time step was successful, false otherwise.
@param[out] pRetMessage Pointer to an initialized string. Messages generated in this function will be appended to the string.
*/
bool CVX_Sim::TimeStep(std::string* pRetMessage)
{
	if(SelfColEnabled) UpdateCollisions(); //update self intersection lists if necessary
	Integrate(CurIntegrator);
	return UpdateStats(pRetMessage);
}

bool CVX_Sim::UpdateStats(std::string* pRetMessage)
{
	//update the overall statisics (can't do this within threaded loops and be safe without mutexes...
	vfloat tmpMaxVoxDisp = 0;
	vfloat tmpMaxVoxVel = 0;
	vfloat tmpMaxBondStrain = 0;
	vfloat tmpMaxBondStress = 0;
	Vec3D tmpTotalObjForce = Vec3D(0,0,0); //reset forces to get new totals
	Vec3D tmpTotalObjDisp = Vec3D(0,0,0);


//TODO: use std::vector iterators!! faster than indexes, for sure!!!!

	int iT = NumVox();
	for (int i=0; i<iT; i++){
		if ((VoxArray[i].S.Pos-VoxArray[i].NomPos).Length() > tmpMaxVoxDisp) tmpMaxVoxDisp = (VoxArray[i].S.Pos-VoxArray[i].NomPos).Length();
		if (VoxArray[i].S.Vel.Length() > tmpMaxVoxVel) tmpMaxVoxVel = VoxArray[i].S.Vel.Length();
		if (VoxArray[i].CurMaxStrain > tmpMaxBondStrain) tmpMaxBondStrain = VoxArray[i].CurMaxStrain;
		if (VoxArray[i].CurMaxStress > tmpMaxBondStress) tmpMaxBondStress = VoxArray[i].CurMaxStress;

		tmpTotalObjForce += VoxArray[i].dS.Force.Abs(); //keep track of forces on global object
		tmpTotalObjDisp += VoxArray[i].S.Vel.Abs()*dt; //keep track of displacements on global object
	}

	int NumberVox = NumVox();
	tmpTotalObjForce /= NumberVox; //normalize...
	tmpTotalObjDisp /= NumberVox;

	MaxVoxDisp = tmpMaxVoxDisp;
	MaxVoxVel = tmpMaxVoxVel;
	MaxBondStrain = tmpMaxBondStrain;
	MaxBondStress = tmpMaxBondStress;
	TotalObjForce = tmpTotalObjForce; //reset forces to get new totals
	TotalObjDisp = tmpTotalObjDisp;
	NormObjForce = tmpTotalObjForce.Length();
	NormObjDisp = tmpTotalObjDisp.Length();



	//Check for divergance...
	if (MaxBondStrain>100){ 
		if (pRetMessage) *pRetMessage = "Simulation Diverged. Please reduce forces or accelerations.\n";
		return false;
	}

	if (pRetMessage) *pRetMessage = "Time step succeeded\n";
	return true;
}


vfloat CVX_Sim::CalcMaxDt(void)
{
	vfloat MaxFreq2 = 0; //maximum frequency in the simulation
	vfloat MaxRFreq2 = 0; //maximum frequency in the simulation

	//std::vector<int>::iterator iBondArray;
	//for (iBondArray=BondArray.begin(); iBondArray!=BondArray.end(); iBondArray++){
	//	if (iBondArray->a1/iBondArray->Vox1()->Mass > MaxFreq2) MaxFreq2 = iBondArray->a1/iBondArray->Vox1()->Mass;
	//	if (iBondArray->a1/iBondArray->Vox2()->Mass > MaxFreq2) MaxFreq2 = iBondArray->a1/iBondArray->Vox2()->Mass;
	//}

	int iT = NumBond();
	for (int i=0; i<iT; i++){
		if (BondArray[i].a1/BondArray[i].Vox1()->Mass > MaxFreq2) MaxFreq2 = BondArray[i].a1/BondArray[i].Vox1()->Mass;
		if (BondArray[i].a1/BondArray[i].Vox2()->Mass > MaxFreq2) MaxFreq2 = BondArray[i].a1/BondArray[i].Vox2()->Mass;
	}

	//calculate dt: (as large as possible...)
	vfloat MaxFreq = sqrt(MaxFreq2);
	return DtFrac/(MaxFreq*2*(vfloat)3.1415926); //safety factor of 10% on the timestep with DtFrac = 0.9...

}

void CVX_Sim::UpdateCollisions(void) // Called every timestep to watch for collisions
{
	//self intersection accumulator
	MaxDispSinceLastBondUpdate += abs(MaxVoxVel*dt/LocalVXC.GetLatticeDim());

	if (CurColSystem == COL_BASIC_HORIZON || CurColSystem == COL_SURFACE_HORIZON){
		if (MaxDispSinceLastBondUpdate > (CollisionHorizon-1.0f)/2){ //if we want to check for self intersection (slow!)
			CalcL1Bonds(CollisionHorizon); //doesn't need to be done every time...
			MaxDispSinceLastBondUpdate = 0.0;
		}
	}
	else{ //if COL_BASIC || COL_SURFACE
		CalcL1Bonds(CollisionHorizon);
	}
}


void CVX_Sim::Integrate(IntegrationType Integrator)
{
	switch (Integrator){
		case I_EULER: {
			//Update Forces...
			int iT = NumBond();
			for (int i=0; i<iT; i++){
				BondArray[i].CalcForce();
//				out << "Bond " << i << "\t" << BondArray[i].Force1.x << "\t" << BondArray[i].Force1.y << "\t" << BondArray[i].Moment1.z << "\t" << BondArray[i].Moment2.z << "\t";
			}

			dt = CalcMaxDt();

			//Update positions... need to do thises seperately if we're going to do damping in the summing forces stage.
			iT = NumVox();
			for (int i=0; i<iT; i++) {
				VoxArray[i].EulerStep();
			//	out << "Vox " << i << "\t" << VoxArray[i].S.Pos.x << "\t" << VoxArray[i].S.Pos.y << "\t" << VoxArray[i].S.Angle.z << "\t";
			//	out << "Vox " << i << "\t" << VoxArray[i].TotVoxForce.x << "\t" << VoxArray[i].TotVoxForce.y << "\t" << VoxArray[i].TotVoxMoment.z << "\t";

			}
			//out << "\n";

		}
		break;

		case I_VERLET: {
			vfloat MaxFreq2 = 0; //maximum frequency in the simulation
			vfloat MaxRFreq2 = 0; //maximum frequency in the simulation

			//Update Forces...
			int iT = NumBond();
			for (int i=0; i<iT; i++){
				BondArray[i].CalcForce();
				if (BondArray[i].a1/BondArray[i].Vox1()->Mass > MaxFreq2) MaxFreq2 = BondArray[i].a1/BondArray[i].Vox1()->Mass;
				if (BondArray[i].a1/BondArray[i].Vox2()->Mass > MaxFreq2) MaxFreq2 = BondArray[i].a1/BondArray[i].Vox2()->Mass;
			}

			//calculate dt: (as large as possible...)
			vfloat MaxFreq = sqrt(MaxFreq2);
			dt = DtFrac/(MaxFreq*2*(vfloat)3.1415926); //safety factor of 10% on the timestep...

			//Update positions... need to do thises seperately if we're going to do damping in the summing forces stage.
			iT = NumVox();
			for (int i=0; i<iT; i++){
				VoxArray[i].VerletStep(dt);
			}
			
		}
		break;
		case I_RK4: {
			//resize the array if needed
			int VoxArraySize = VoxArray.size();
			if (OrigVoxArray.size() != VoxArraySize){ //if sizes have changed...
				OrigVoxArray.resize(VoxArraySize);
				A.resize(VoxArraySize);
				B.resize(VoxArraySize);
				C.resize(VoxArraySize);
				D.resize(VoxArraySize);
			}
			int BondArraySize = BondArray.size();
			if (OrigBondArray.size() != BondArraySize)	OrigBondArray.resize(BondArraySize);


			dt = (vfloat)0.00014*DtFrac;
			int i, iT;
			int NV = NumVox(); //num voxels constant, num bonds not necessarily
			//save the initial state.
			for (i=0; i<VoxArraySize; i++) OrigVoxArray[i] = VoxArray[i];
			for (i=0; i<BondArraySize; i++) OrigBondArray[i] = BondArray[i];

			//get the derivative at the current state:
			iT = NumBond();
			for (i=0; i<iT; i++) BondArray[i].CalcForce(); //CalcForce() uses whatever is in VoxArray to calculate forces...
			for (i=0; i<VoxArraySize; i++) A[i] = VoxArray[i]; //set A = to original voxel array
			for (i=0; i<NV; i++) A[i].UpdateDState(); //updates derivative of state A based on the S we have in VoxArray at the end of last time step

			for (i=0; i<NV; i++) VoxArray[i].UpdateState(dt/2, A[i].dS); //Advances VoxArray according to derivative of A
			iT = NumBond();
			for (i=0; i<iT; i++)	BondArray[i].CalcForce(); //CalcForce() uses whatever is in VoxArray to calculate forces...
			for (i=0; i<VoxArraySize; i++) B[i] = VoxArray[i]; //set B = to original voxel array
			for (i=0; i<NV; i++) B[i].UpdateDState(); //updates derivate of B at 1/2 dt
			VoxArray = OrigVoxArray; //reset VoxArray

			for (i=0; i<NV; i++) VoxArray[i].UpdateState(dt/2, B[i].dS);
			iT = NumBond();
			for (i=0; i<iT; i++) BondArray[i].CalcForce(); //CalcForce() uses whatever is in VoxArray to calculate forces...
			for (i=0; i<VoxArraySize; i++) C[i] = VoxArray[i]; //set C = to original voxel array
			for (i=0; i<NV; i++) C[i].UpdateDState();
			VoxArray = OrigVoxArray; //reset VoxArray

			for (int i=0; i<NV; i++) VoxArray[i].UpdateState(dt, C[i].dS);
			iT = NumBond();
			for (int i=0; i<iT; i++)	BondArray[i].CalcForce(); //CalcForce() uses whatever is in VoxArray to calculate forces...
			for (int i=0; i<VoxArraySize; i++) D[i] = VoxArray[i]; //set D = to original voxel array
			for (int i=0; i<NV; i++) D[i].UpdateDState();
			VoxArray = OrigVoxArray; //reset VoxArray

			//do the RK4 magic!
			for (int i=0; i<NV; i++){
				VoxArray[i].dS.Force = (A[i].dS.Force + 2.0*(B[i].dS.Force + C[i].dS.Force) + D[i].dS.Force)/6.0;
				VoxArray[i].dS.LinMom = (A[i].dS.LinMom + 2.0*(B[i].dS.LinMom + C[i].dS.LinMom) + D[i].dS.LinMom)/6.0;
				VoxArray[i].dS.Moment = (A[i].dS.Moment + 2.0*(B[i].dS.Moment + C[i].dS.Moment) + D[i].dS.Moment)/6.0;
				VoxArray[i].dS.AngMom = (A[i].dS.AngMom + 2.0*(B[i].dS.AngMom + C[i].dS.AngMom) + D[i].dS.AngMom)/6.0;
				VoxArray[i].dS.ScaleForce = (A[i].dS.ScaleForce + 2.0*(B[i].dS.ScaleForce + C[i].dS.ScaleForce) + D[i].dS.ScaleForce)/6.0;
				VoxArray[i].dS.ScaleMom = (A[i].dS.ScaleMom + 2.0*(B[i].dS.ScaleMom + C[i].dS.ScaleMom) + D[i].dS.ScaleMom)/6.0;
			}
			for (int i=0; i<NV; i++) VoxArray[i].UpdateState(dt, VoxArray[i].dS);


		}
		break;

	}
}

#ifdef USE_OPEN_GL

void CVX_Sim::Draw(int Selected, bool ViewSection, int SectionLayer)
{
	if (!Initalized) return;

	if (pEnv->FloorEnabled) DrawFloor(); //draw the floor if its in use
	if (ViewBonds) DrawBonds();
	if (ViewGeo) DrawGeometry(Selected, ViewSection, SectionLayer); //section view only currently enabled in voxel view mode
	if (ViewSurfMesh) DrawSurfMesh(Selected);
	if (ViewGeoMesh) DrawVoxMesh(Selected);
	if (ViewAngles)	DrawAngles();

}

void CVX_Sim::DrawFloor(void)
{

	//TODO: build an openGL list 
	vfloat Size = LocalVXC.GetLatticeDim()*4;
	vfloat sX = 1.5*Size;
	vfloat sY = .866*Size;

	glEnable(GL_LIGHTING);

	glLoadName (-1); //never want to pick floor

	glNormal3d(0.0, 0.0, 1.0);
	for (int i=-20; i <=30; i++){
		for (int j=-40; j <=60; j++){
			glColor4d(0.6, 0.7+0.2*((int)(1000*sin((float)(i+110)*(j+106)*(j+302)))%10)/10.0, 0.6, 1.0);
			glBegin(GL_TRIANGLE_FAN);
			glVertex3d(i*sX, j*sY, 0.0);
			glVertex3d(i*sX+0.5*Size, j*sY, 0.0);
			glVertex3d(i*sX+0.25*Size, j*sY+0.433*Size, 0.0);
			glVertex3d(i*sX-0.25*Size, j*sY+0.433*Size, 0.0);
			glVertex3d(i*sX-0.5*Size, j*sY, 0.0);
			glVertex3d(i*sX-0.25*Size, j*sY-0.433*Size, 0.0);
			glVertex3d(i*sX+0.25*Size, j*sY-0.433*Size, 0.0);
			glVertex3d(i*sX+0.5*Size, j*sY, 0.0);
			glEnd();

			glColor4d(0.6, 0.7+0.2*((int)(1000*sin((float)(i+100)*(j+103)*(j+369)))%10)/10.0, 0.6, 1.0);
			
			glBegin(GL_TRIANGLE_FAN);
			glVertex3d(i*sX+.75*Size, j*sY+0.433*Size, 0.0);
			glVertex3d(i*sX+1.25*Size, j*sY+0.433*Size, 0.0);
			glVertex3d(i*sX+Size, j*sY+0.866*Size, 0.0);
			glVertex3d(i*sX+0.5*Size, j*sY+0.866*Size, 0.0);
			glVertex3d(i*sX+0.25*Size, j*sY+0.433*Size, 0.0);
			glVertex3d(i*sX+0.5*Size, j*sY, 0.0);
			glVertex3d(i*sX+Size, j*sY, 0.0);
			glVertex3d(i*sX+1.25*Size, j*sY+0.433*Size, 0.0);
			glEnd();
		}
	}
}

void CVX_Sim::DrawGeometry(int Selected, bool ViewSection, int SectionLayer)
{
	Vec3D Center;

	int iT = NumVox();
	int x, y, z;
	CColor ThisColor;
	for (int i = 0; i<iT; i++) //go through all the voxels...
	{
		pEnv->pObj->GetXYZNom(&x, &y, &z, StoXIndexMap[i]);
		if (ViewSection && z>SectionLayer) continue; //exit if obscured in a section view!


		Center = VoxArray[i].S.Pos;

		ThisColor = GetCurColor(i, Selected);
		glColor4d(ThisColor.r, ThisColor.g, ThisColor.b, ThisColor.a);
		

		Vec3D tmp = Vec3D(0,0,0);
		glPushMatrix();
		glTranslated(Center.x, Center.y, Center.z);

		glLoadName (StoXIndexMap[i]); //to enable picking

		//generate rotation matrix here!!! (from quaternion)
		Vec3D Axis;
		vfloat AngleAmt;
		VoxArray[i].S.Angle.AngleAxis(AngleAmt, Axis);
		glRotated(AngleAmt*180/3.1415926, Axis.x, Axis.y, Axis.z);
	
		double TempFact = 1.0;
		if(pEnv->TempEnabled) TempFact = (1+(LocalVXC.GetBaseMat(VoxArray[i].GetMatIndex())->GetMatTemp() - pEnv->TempBase) * LocalVXC.GetBaseMat(VoxArray[i].GetMatIndex())->GetCTE());

		Vec3D Scale = TempFact*VoxArray[i].S.Scale.ScaleInv(VoxArray[i].NomScale); //show deformed voxel size
		glScaled(Scale.x, Scale.y, Scale.z);

		//LocalVXC.Voxel.DrawVoxel(&tmp, LocalVXC.Lattice.Lattice_Dim*(1+0.5*CurTemp * pMaterials[CVoxelArray[i].MatIndex].CTE), LocalVXC.Lattice.Z_Dim_Adj);
		LocalVXC.Voxel.DrawVoxel(&tmp, LocalVXC.GetLatticeDim()); //[i].CurSize.x); //, LocalVXC.Lattice.Z_Dim_Adj);
		
		glPopMatrix();
	}
}

CColor CVX_Sim::GetCurColor(int SIndex, int Selected)
{
	if (StoXIndexMap[SIndex] == Selected) return CColor(1.0f, 0.0f, 1.0f, 1.0f); //highlight selected voxel (takes precedence...)

	switch (CurViewCol) {
		case RVC_DISP:
			return GetJet((VoxArray[SIndex].S.Pos - VoxArray[SIndex].NomPos).Length() /MaxVoxDisp);
			break;
		case RVC_STRAIN:
			return GetJet(VoxArray[SIndex].CurMaxStrain / MaxBondStrain);
			break;
		case RVC_STRESS:
			return GetJet(VoxArray[SIndex].CurMaxStress / MaxBondStress);
			break;
		case RVC_MATERIAL:
			float R, G, B, A;
			LocalVXC.GetLeafMat(VoxArray[SIndex].MyXIndex)->GetColorf(&R, &G, &B, &A);
			return CColor(R, G, B, A);
			break;
		case RVC_STATE:
			if (VoxArray[SIndex].VBroken) return CColor(1.0f, 0.0f, 0.0f, 1.0f);
			else if (VoxArray[SIndex].VYielded) return CColor(1.0f, 1.0f, 0.0f, 1.0f);
			else return CColor(1.0f, 1.0f, 1.0f, 1.0f);
			break;
		default:
			return CColor(1.0f,1.0f,1.0f, 1.0f);
			break;
	}

	
}

void CVX_Sim::DrawSurfMesh(int Selected)
{
	SurfMesh.UpdateMesh(Selected); //updates the generated mesh
	SurfMesh.Draw();
}

void CVX_Sim::DrawVoxMesh(int Selected)
{
	VoxMesh.UpdateMesh(Selected); //updates the generated mesh
	VoxMesh.Draw();
}


void CVX_Sim::DrawBonds(void)
{
	Vec3D P1;
	Vec3D P2;
	float ColFac = 0.25;

	float PrevLineWidth;
	glGetFloatv(GL_LINE_WIDTH, &PrevLineWidth);
	glLineWidth(2.0);

	glDisable(GL_LIGHTING);
	float adraw = 0.5f;
	

	glBegin(GL_LINES);

	int iT = NumBond();
	for (int i = 0; i<iT; i++) //go through all the bonds...
	{
		P1 = BondArray[i].Vox1()->S.Pos;
		P2 = BondArray[i].Vox2()->S.Pos;
		vfloat Force = BondArray[i].Force1.Length(); //check which force to use!
		bool DrawMe = true;

		switch (BondArray[i].ThisBondType){
			case B_LINEAR:
				glColor4d(0.0,0.0,0.0, adraw);
			break;

			case B_LINEAR_CONTACT:
				if (Force == 0.0) glColor4f(0.3f,0.3f,1.0f, adraw);
				else glColor4f(1.0f,0.0f,0.0f, adraw);
			break;

			case B_TILE_EDGE:
//				if ((*(BondArray+i)).Yielded) glColor4f(255, 0, 255, adraw);
//				else
				glColor4f(ColFac,ColFac,(float)(ColFac+(1-ColFac)*Force), adraw);
			break;

			case B_TILE_ZCORNER:
//				if ((*(BondArray+i)).Yielded) glColor4f(0, 255, 0, adraw);
//				else
				glColor4f(ColFac, (float)(ColFac+(1-ColFac)*Force), (float)(ColFac+(1-ColFac)*Force), adraw);
			break;
			case B_INPUT_LINEAR_NOROT:
				glColor4f(1.0, 0.0, 0.0, adraw);
				if (!Dragging) DrawMe = false;
			break;

		}

		//draw the line!
		if (DrawMe) {glVertex3f((float)P1.x, (float)P1.y, (float)P1.z); glVertex3f((float)P2.x, (float)P2.y, (float)P2.z);}

	}
	
	glEnd();

	//draw triangle for points that are stuck via static friction
	glBegin(GL_TRIANGLES);
	glColor4f(255, 255, 0, adraw);
	vfloat dist = VoxArray[0].NomScale.x/3; //needs work!!
	iT = NumVox();
	for (int i = 0; i<iT; i++){ //go through all the voxels...
		if (VoxArray[i].StaticFricFlag){ //draw point if static friction...
			P1 = VoxArray[i].S.Pos;
			glVertex3f((float)P1.x, (float)P1.y, (float)P1.z); 
			glVertex3f((float)P1.x, (float)(P1.y - dist/2), (float)(P1.z + dist));
			glVertex3f((float)P1.x, (float)(P1.y + dist/2), (float)(P1.z + dist));
		}
	}
	glEnd();

	glLineWidth(PrevLineWidth);
	glEnable(GL_LIGHTING);

}

void CVX_Sim::DrawAngles(void)
{
	//draw directions
	float PrevLineWidth;
	glGetFloatv(GL_LINE_WIDTH, &PrevLineWidth);
	glLineWidth(2.0);

	glBegin(GL_LINES);

	for (int i = 0; i<NumVox(); i++){ //go through all the voxels... (GOOD FOR ONLY SMALL DISPLACEMENTS, I THINK... think through transformations here!)
		glColor3f(1,0,0); //+X direction
		glVertex3d(VoxArray[i].S.Pos.x, VoxArray[i].S.Pos.y, VoxArray[i].S.Pos.z);
		Vec3D Axis1 = Vec3D(LocalVXC.GetLatticeDim()/4,0,0);
		Vec3D RotAxis1 = (VoxArray[i].S.Angle*CQuat(Axis1)*VoxArray[i].S.Angle.Conjugate()).ToVec();
		glVertex3d(VoxArray[i].S.Pos.x + RotAxis1.x, VoxArray[i].S.Pos.y + RotAxis1.y, VoxArray[i].S.Pos.z + RotAxis1.z);

		glColor3f(0,1,0); //+Y direction
		glVertex3d(VoxArray[i].S.Pos.x, VoxArray[i].S.Pos.y, VoxArray[i].S.Pos.z);
		Axis1 = Vec3D(0, LocalVXC.GetLatticeDim()/4,0);
		RotAxis1 = (VoxArray[i].S.Angle*CQuat(Axis1)*VoxArray[i].S.Angle.Conjugate()).ToVec();
		glVertex3d(VoxArray[i].S.Pos.x + RotAxis1.x, VoxArray[i].S.Pos.y + RotAxis1.y, VoxArray[i].S.Pos.z + RotAxis1.z);

		glColor3f(0,0,1); //+Z direction
		glVertex3d(VoxArray[i].S.Pos.x, VoxArray[i].S.Pos.y, VoxArray[i].S.Pos.z);
		Axis1 = Vec3D(0,0, LocalVXC.GetLatticeDim()/4);
		RotAxis1 = (VoxArray[i].S.Angle*CQuat(Axis1)*VoxArray[i].S.Angle.Conjugate()).ToVec();
		glVertex3d(VoxArray[i].S.Pos.x + RotAxis1.x, VoxArray[i].S.Pos.y + RotAxis1.y, VoxArray[i].S.Pos.z + RotAxis1.z);

	}
	glEnd();

	glLineWidth(PrevLineWidth);

	glEnable(GL_LIGHTING);
}



#endif //OPENGL

void CVX_Sim::CalcL1Bonds(vfloat Dist) //creates contact bonds for all voxels within specified distance
{
//	Dist = Dist*LocalVXC.GetLatticeDim();
	vfloat Dist2;

	DeleteTmpBonds();

	if (CurColSystem == COL_SURFACE || COL_SURFACE_HORIZON){
		//clear bond array past the m_NumBonds position
		for (int i=0; i<NumSurfVoxels(); i++){ //go through each combination of surface voxels...
			int SIndex1 = SurfVoxels[i];

			for (int j=i+1; j<NumSurfVoxels(); j++){
				int SIndex2 = SurfVoxels[j];
				if (!VoxArray[SIndex1].IsNearbyVox(SIndex2)){
					vfloat ActDist = Dist*(VoxArray[SIndex1].S.Scale.x + VoxArray[SIndex1].S.Scale.x)/2; //ASSUMES ISOTROPIC!!

					Dist2 = (VoxArray[SIndex1].S.Pos - VoxArray[SIndex2].S.Pos).Length2();
					if (Dist2 < ActDist*ActDist){ //if within the threshold
						CreateBond(B_LINEAR_CONTACT, SIndex1, SIndex2); //create temporary bond...	
					}

				}
			}
		}
	}
	else { //check against all!
		for (int i=0; i<NumVox(); i++){ //go through each combination of voxels...
			int SIndex1 = i;

			for (int j=i+1; j<NumVox(); j++){
				int SIndex2 = j;
				if (!VoxArray[SIndex1].IsNearbyVox(SIndex2)){

					vfloat ActDist = Dist*(VoxArray[SIndex1].S.Scale.x + VoxArray[SIndex1].S.Scale.x)/2; //ASSUMES ISOTROPIC!!
					Dist2 = (VoxArray[SIndex1].S.Pos - VoxArray[SIndex2].S.Pos).Length2();
					if (Dist2 < ActDist*ActDist){ //if within the threshold
						CreateBond(B_LINEAR_CONTACT, SIndex1, SIndex2); //create temporary bond...	
					}

				}
			}
		}
	}
}

Vec3D CVX_Sim::GetCM(void)
{

	vfloat TotalMass = 0;
	Vec3D Sum = Vec3D(0,0,0);
	int iT = NumVox();
	for (int i=0; i<iT; i++){
		Sum += VoxArray[i].S.Pos*VoxArray[i].Mass;
		TotalMass += VoxArray[i].Mass;
	}

	return Sum/TotalMass;
}

Vec3D CVX_Sim::GetSumForce(CVX_FRegion* pRegion)
{
	Vec3D ForceSum = Vec3D(0,0,0);

	Vec3D BCpoint;
	Vec3D BCsize = pEnv->pObj->GetLatDimEnv()/2.0;
	Vec3D WSSize = pEnv->pObj->GetWorkSpace();

	int NumVoxS = NumVox();
	for (int i=0; i<NumVoxS; i++){
		BCpoint = pEnv->pObj->GetXYZ(StoXIndexMap[i]);
		if (pRegion->pRegion->IsTouching(&BCpoint, &BCsize, &WSSize)){
			ForceSum += VoxArray[i].CalcTotalForce(false);
		}
	}

	return ForceSum;
}

vfloat CVX_Sim::GetSumForceDir(CVX_FRegion* pRegion)
{
	//right now only fixed regions... (forced regions should be zero!)
	//get force only in dircetion of pull!
	Vec3D Res = GetSumForce(pRegion);

	Vec3D Dir = pRegion->Displace;
	if (Dir.Length2() == 0) return Res.Length(); //return magnitude of no direction...
	else {
		Dir.Normalize();
		return Res.Dot(Dir);
	}

}

int CVX_Sim::NumYielded(void)
{
	int NumYieldRet = 0;
	int NumBondNow = (int)(BondArray.size());
	for (int i=0; i<NumBondNow; i++)
		if (BondArray[i].Yielded) NumYieldRet++;

	return NumYieldRet;
}

int CVX_Sim::NumBroken(void)
{
	int NumBrokenRet = 0;
	int NumBondNow = (int)(BondArray.size());
	for (int i=0; i<NumBondNow; i++)
		if (BondArray[i].Broken) NumBrokenRet++;

	return NumBrokenRet;

}