/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#ifndef VX_SIM_H
#define VX_SIM_H

#include "VXS_Voxel.h"
#include "VXS_Bond.h"
#include "VX_Environment.h"
#include "VX_MeshUtil.h"

#ifdef USE_OPEN_GL
#include "MarchCube.h"
#include "Array3D.h"
#ifdef QT_GUI_LIB
#include <qgl.h>
#else
#include "OpenGLInclude.h" //If not using QT's openGL system, make a header file "OpenGLInclude.h" that includes openGL library functions 
#endif
#endif

//! Defiens the color to display each voxel
enum ViewColor {
	RVC_DISP, //!< Color coded displacement from inital position.
	RVC_STRAIN, //!< Color coded engineering srain, or internal deformation percent.
	RVC_STRESS, //!< Color coded engineerign stress, or internal force
	RVC_MATERIAL, //!< Material color
	RVC_STATE //Color coded information about the state of each voxel (reserved for future implementation)
};
//! The integrator to be used to advance the simulation each timestep.
enum IntegrationType {
	I_EULER, //!< Euler integration. This should be suitable for nearly all situations. 
	I_RK4, //!< Runge-Kutta integration. Slower than Euler with benefits only when stiffnesses change quickly.
	I_VERLET //!< Verlet integration. Similar performance to Euler but without explicit momentum.
};
//! The collision detection scheme to use if enabled.
enum ColSystem{
	COL_BASIC, //!< Basic n2 collision detection. Extremely slow.
	COL_SURFACE, //!< Basic n2 collision detection, but only considering voxels on the surface of the object. Very slow.
	COL_BASIC_HORIZON, //!< Hierarchical collision detection between all voxels. (One level) Updates potential collision list only when aggregated motion requires it.
	COL_SURFACE_HORIZON //!< Hierarchical collision detection between surface voxels. (One level) This is the fastest collision scheme implemented, and suffice for most cases. Updates potential collision list only when aggregated motion requires it.
};

enum MatBlendModel{
	MB_LINEAR, //blended materials combine stiffness linearly (x)
	MB_EXPONENTIAL, //blended materials combine stiffness exponentially (2^x-1) (-1 puts in Y range of 0 to 1
	MB_POLYNOMIAL //blended materials combine stiffness polynomially (x^n) with n stored seperately
};

enum Axis {XAXIS, YAXIS, ZAXIS}; //which axis do we refer to?
#define MIN_TEMP_FACTOR 0.1 /*The smallest a voxel can go.*/

//!Dynamic simulation class for time simulation of voxel objects.
/*!
To enable openGL rendering functions, define USE_OPEN_GL somewhere in the compiler's pre-processing routine.
*/
class CVX_Sim
{
public:
	CVX_Sim(void); //!< Constructor
	~CVX_Sim(void); //!< Destructor
	CVX_Sim& operator=(const CVX_Sim& rSim); //!< Overload "=" 

//	std::ofstream out;

	//I/O function for save/loading
	void SaveVXAFile(std::string filename);
	void SaveResultFile(std::string filename);
	bool LoadVXAFile(std::string filename);

	void WriteVXA(CXML_Rip* pXML);
	void WriteResultFile(CXML_Rip* pXML);
	bool ReadVXA(CXML_Rip* pXML, std::string* RetMessage = NULL);

	//I/O function for save/loading
	void WriteXML(CXML_Rip* pXML);
	bool ReadXML(CXML_Rip* pXML, std::string* RetMessage = NULL);

	//input information
	CVX_Environment* pEnv; //!< Pointer to the physical environment information. This variable is set on import() and should not be manually changed.
	CVX_Object LocalVXC; //!< Local copy of the voxel object. This copy is stored to ensure it never changes throughout the simulation. This variable is set on import() and should not be manually changed.

	//Simulation information
	std::vector<CVXS_Voxel> VoxArray; //!< The main array of voxels.
	int NumVox(void); //inline int NumVox(void); //!< Returns the number of voxels in the simulation.
	std::vector<CVXS_Bond> BondArray; //!< The main array of bonds.
	int NumBond(void); //inline int NumBond(void); //!< Returns the number of bonds in the simulation.
	float	SimStopTime;	//!<Used to specify the amount of time that a simulation should run (seconds). Simulation will run until this value or SimStopStep is exceeded
	long int	SimStopStep;	//!<Used to specify the number of steps that a simulation should run. Simulation will run until this value or SimStopTime is exceeded
	float	Fitness;	//!<Keeps track of whatever fitness we choose to track
	int	FitnessRepType; //!<Holds the fitness reporting type. For now =0 tracks the center of mass, =1 tracks a particular Voxel number
	int	TrackVoxel;		//!<Holds the particular voxel that will be tracked (if used).
	std::string FitnessFileNm;	//!<Holds the filename of the fitness output file that might be used
	bool WriteFitnessFile;
	bool print_scrn;	//!<flags whether status will be sent to the console

	std::vector<int> XtoSIndexMap; //!< Maps the global CVX_Object index to the corresponding CVX_Sim voxel index.
	std::vector<int> StoXIndexMap; //!< Maps CVX_Sim voxel index to the original global CVX_Object index.
	int GetVoxIndex(int i, int j, int k) {return XtoSIndexMap[LocalVXC.GetIndex(i, j, k)];} //!< Returns the CVX_SIM voxel index at specified voxel location. If there is no instantiated voxel here -1 is returned. @param[in] i The X Voxel index of the desired voxel. @param[in] j The Y Voxel index of the desired voxel. @param[in] k The Z Voxel index of the desired voxel.

	//Simulation Management
	void Import(CVX_Environment* pEnvIn = NULL, CMesh* pSurfMeshIn = NULL, std::string* RetMessage = NULL); //!< Imports a physical environment into the simulator.
	bool CreateBond(BondType BondTypeIn, int SIndex1, int SIndex2, bool Perm = false); //!< Creates a new bond between two voxels. 
	void DeleteTmpBonds(void); //!< Deletes all non-permanent bonds.
	bool IsInitalized(void) {return Initalized;} //!< Returns true if a valid environment has been imported.

	void ClearAll(void); //!< Clears all environment-specific information form the simulation.
	void ResetSimulation(void); //!< Resets the environment to its initial imported state.

	//Integration 
	bool TimeStep(std::string* pRetMessage = NULL); //!< Advances the simulation one time step.
	vfloat CalcMaxDt(void); //!< Calculates the current maximum timestep based on the highest resonant frequency in the object.
	vfloat DtFrac; //percent of maximum dt to use
	vfloat dt; //seconds per timestep

	//Simulator features:
	//Self collision:
	void EnableSelfCollision(bool Enabled) {SelfColEnabled = Enabled;} //!< Enables or disables self-collision in the simulation.
	bool IsSelfColEnabled(void) {return SelfColEnabled;}
	std::vector<int> SurfVoxels; //A list of voxels that are on the surface (IE eligible for contact bonds...) (containts SIndex!)
	int NumSurfVoxels(void) {return (int)SurfVoxels.size();}; //how 
	void CalcL1Bonds(vfloat Dist); //creates contact bonds for all voxels within specified distance
	vfloat MaxDispSinceLastBondUpdate;

	//Damping:
	void SetBondDampZ(vfloat BondDampZIn) {BondDampingZ = BondDampZIn;} //!< Sets the damping ratio for connected voxels. When this is non-zero, each voxel is damped (based on its mass and stiffness) according to its relative velocity to the other voxel in each bond. Range is [0.0 to 1.0]. Values greater than 1.0 may cause numerical instability.
	void SetCollisionDampZ(vfloat ColDampZIn) {ColDampingZ = ColDampZIn;} //!< Sets the damping ratio for voxels in colliding state. When this is non-zero, each voxel is damped (based on its mass and stiffness) according to the penetration velocity. Range is [0.0 to 1.0]. Values greater than 1.0 may cause numerical instability.
	void SetSlowDampZ(vfloat SlowDampIn) {SlowDampingZ = SlowDampIn;} //!< Sets the damping ratio that slows downs voxels. When this is non-zero, each voxel is damped (based on its mass and stiffness) to ground. Range is [0.0 to 1.0]. Values greater than 1.0 may cause numerical instability.
	
	vfloat GetBondDampZ(void) {return BondDampingZ;} //!< Returns the current bond damping.
	vfloat GetCollisionDampZ(void) {return ColDampingZ;} //!< Returns the current collision damping.
	vfloat GetSlowDampZ(void) {return SlowDampingZ;} //!< Returns the current voxel slowing damping.

	//Plasticity:
	void EnablePlasticity(bool Enabled) {PlasticityEnabled = Enabled;}
	bool IsPlasticityEnabled(void) {return PlasticityEnabled;}
	void EnableFailure(bool Enabled) {FailureEnabled = Enabled;}
	bool IsFailureEnabled(void) {return FailureEnabled;}
	int NumYielded(void);
	int NumBroken(void);

	//Material Blending
	bool BlendingEnabled;
	vfloat MixRadius;
	MatBlendModel BlendModel;
	vfloat PolyExp; //polynomial exponent if using polynomial model

//	void SetViscousDamp(double ViscDampIn) {ViscousDamp = ViscDampIn;}
//	double ViscousDamp; //0->1. percent of velocity to shave off per second
	bool FluidDampEnabled; //do we want fluid-style damping?
	
	//Input voxel (moved by user interactively or otherwise
	bool Dragging; //Is the input voxel active
	int CurXSel; //current selection... (VXC index...)
	int InputVoxSInd;
	int InputBondInd;
	CVXS_Voxel* InputVoxel(void);
	CVXS_Bond* InputBond(void);
	Vec3D PointerOffset;

	//Do some pseudo poisson's ratio constraints? (not useful yet)
	bool PoissonKickBackEnabled; //should the aspect ratio of the voxels affect the forces?

	//Impose angle and extension limits on the bonds (currently unstable)
	bool EnforceLatticeEnabled; //have limits for local bedning.compressive angles? (try again with RKS... a bit unstable with Euler du to rapidly changing stiffness

	void EnableMaxVelLimit(bool Enabled) {MaxVelLimitEnabled=Enabled;}
	bool IsMaxVelLimitEnabled(void) {return MaxVelLimitEnabled;}
	vfloat GetMaxVoxVelLimit(void) {return MaxVoxVelLimit;}
	void SetMaxVoxVelLimit(vfloat NewMaxVoxVelLimit) {MaxVoxVelLimit = NewMaxVoxVelLimit;}

	//Information about current state:
	Vec3D TotalObjForce; //a vector total of the magnitude of forces on object
	vfloat NormObjForce; //reduced to a scalar (magnitude) 
	Vec3D TotalObjDisp; //a vector total of the magnitude of displacements
	vfloat NormObjDisp; //reduced to a scalar (magnitude) 
	vfloat MaxVoxDisp, MaxVoxVel, MaxBondStrain, MaxBondStress;

	Vec3D GetCM(void);
	Vec3D IniCM; //initial center of mass
	Vec3D CurCM;

	Vec3D GetSumForce(CVX_FRegion* pRegion); //returns total force on a region
	vfloat GetSumForceDir(CVX_FRegion* pRegion);

protected:
	bool Initalized; //!< Flag to denote if simulation is runnable. True if there is an environement successfully loaded, false otherwise.

	//Integration
	void Integrate(IntegrationType Integrator = I_EULER);
	bool UpdateStats(std::string* pRetMessage); //returns false if simualtion diverged...
	IntegrationType CurIntegrator;
	//States for RK4 to use without re-initializing memory every time step...
	std::vector<CVXS_Voxel> OrigVoxArray, A, B, C, D;
	std::vector<CVXS_Bond> OrigBondArray;

	//collision
	void UpdateCollisions(void); //!< Called every timestep to watch for collisions
	bool SelfColEnabled; //do we want the overhead of watching for self-intersections? (for large displacement systems)
	ColSystem CurColSystem;
	vfloat CollisionHorizon; //multiple of voxel dimension to add to potential collision bond list

	//Damping:
	vfloat BondDampingZ; //Damping factor zeta (1 = critical damping, 0 = no damping...
	vfloat ColDampingZ;	//Damping factor zeta (1 = critical damping, 0 = no damping...
	vfloat SlowDampingZ; //Damping factor zeta (1 = critical damping, 0 = no damping...

	//Plasticity:
	bool PlasticityEnabled; //enables plastic deformation of non-linear materials
	bool FailureEnabled; //enables failure of materials (breaking of bonds)

	//Material Blending:


	//Velocity limit
	bool MaxVelLimitEnabled;
	vfloat MaxVoxVelLimit; //in percentage of voxel base size per time step (1 = can move whole voxel distance in a single timestep)

	public:
	CVX_MeshUtil SurfMesh; //local copy of any imported geometry surface mesh (optional)

#ifdef USE_OPEN_GL
	bool DrawView;
	bool ViewGeo; //look at geometry?
	bool ViewGeoMesh; //if looking at geometry, do we want the mesh?
	bool ViewSurfMesh; //do we want to see imported surface mesh (if exists)
	bool ViewBonds; //look at Bonds?
	bool ViewForce; //look at force vectors?
	bool ViewAngles; //look at axes for each point?

	CVX_MeshUtil VoxMesh; //voxel geometry mesh (generated internally)
	ViewColor CurViewCol;
	CColor GetJet(vfloat val){if (val > 0.75) return CColor(1, 4-val*4, 0, 1.0);	else if (val > 0.5) return CColor(val*4-2, 1, 0, 1.0); else if (val > 0.25) return CColor(0, 1, 2-val*4, 1.0); else return CColor(0, val*4, 1, 1.0);};
	CColor GetCurColor(int SIndex, int Selected);

	//Drawing
	void Draw(int Selected = -1, bool ViewSection=false, int SectionLayer=0);
	void DrawGeometry(int Selected = -1, bool ViewSection=false, int SectionLayer=0);
	void DrawSurfMesh(int Selected = -1);
	void DrawVoxMesh(int Selected = -1);
	void DrawFloor(void);
	void DrawBonds(void);
	void DrawForce(void);
	void DrawAngles(void);

#endif
};

#endif //VX_SIM_H
