/*******************************************************************************
Copyright (c) 2012, Jonathan Hiller

This file is part of the AMF Tools suite. http://amf.wikispaces.com/
AMF Tools is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
AMF Tools is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#ifndef _VEC3D_H
#define _VEC3D_H

//Possible Linux portability issues: min, max

#include <math.h>
#include <float.h>

//typedef double vfloat; //precision underlying all 3D work...
//typedef double vfloat; //precision underlying all 3D work...
#define vfloat float

class CQuat;
class Vec3D {
public:
	//Constructors
	Vec3D() {x = 0; y = 0; z = 0;}
	//Vec3D(const Vec3D& s) {x = s.x; y = s.y; z = s.z;}
	Vec3D(const Vec3D& Vec) {*this = Vec;} //copy constructor

	Vec3D(const vfloat dx, const vfloat dy, const vfloat dz) {x = dx; y = dy; z = dz;}
	inline Vec3D& operator=(const Vec3D& s) {x = s.x; y = s.y; z = s.z; return *this; }; //overload equals


#ifdef WIN32
	bool IsValid() const {return !_isnan((double)x) && _finite((double)x) && !_isnan((double)y) && _finite((double)y) && !_isnan((double)z) && _finite((double)z);} //is a valid vector? (funky windows _ versions...)
#else
	bool IsValid() const {return !__isnan(x) && finite(x) && !__isnan(y) && finite(y) && !__isnan(z) && finite(z);} //is a valid vector?
#endif

	bool IsNear(const Vec3D& s, const vfloat thresh) {return Dist2(s)<thresh*thresh;}

	//Attributes
	vfloat x, y, z;
	inline vfloat getX(void) const {return x;};
	inline vfloat getY(void) const {return y;};
	inline vfloat getZ(void) const {return z;};
	inline void setX(const vfloat XIn) {x = XIn;};
	inline void setY(const vfloat YIn) {y = YIn;};
	inline void setZ(const vfloat ZIn) {z = ZIn;};


	//these should all be friends....
	//Operators
	friend Vec3D operator+(const Vec3D& v1, const Vec3D& v2) {return Vec3D(v1.x+v2.x, v1.y+v2.y, v1.z+v2.z);} //Plus
	friend Vec3D operator-(const Vec3D& v1, const Vec3D& v2) {return Vec3D(v1.x-v2.x, v1.y-v2.y, v1.z-v2.z);} //Minus
	friend Vec3D operator*(const Vec3D& v, const vfloat f) {return Vec3D(v.x*f, v.y*f, v.z*f);} //Multiply
	friend Vec3D operator*(const vfloat f, const Vec3D& v) {return v*f;}
	friend Vec3D operator/(const Vec3D& v, const vfloat f) {return Vec3D(v.x/f, v.y/f, v.z/f);} //Divide
	friend Vec3D operator-(const Vec3D& v) {return Vec3D(-v.x, -v.y, -v.z);} //Negative

	friend bool operator==(const Vec3D& v1, const Vec3D& v2) {return (v1.x==v2.x && v1.y==v2.y && v1.z==v2.z);} //Is equal
	friend bool operator!=(const Vec3D& v1, const Vec3D& v2) {return !(v1==v2);} //Is not equal
	Vec3D& operator+=(const Vec3D& s) {x += s.x; y += s.y; z += s.z; return *this;} //add and set
	Vec3D& operator-=(const Vec3D& s) {x -= s.x; y -= s.y; z -= s.z; return *this;} //subract and set
	Vec3D& operator*=(const vfloat f) {x *= f; y *= f; z *= f; return *this;} //multiply and set
	Vec3D& operator/=(const vfloat f) {x /= f; y /= f; z /= f; return *this;} //subtract and set
	
	//Vector operations (change this object)
	inline vfloat Normalize() {vfloat l = sqrt(x*x+y*y+z*z); if (l > 0) {x /= l;y /= l;z /= l;} return l;} //normalizes this vector
	inline Vec3D Rot(const Vec3D u, const vfloat a) {vfloat ca = cos(a); vfloat sa = sin(a); vfloat t = 1-ca; return Vec3D((u.x*u.x*t + ca) * x + (u.x*u.y*t - u.z*sa) * y + (u.z*u.x*t + u.y*sa) * z, (u.x*u.y*t + u.z*sa) * x + (u.y*u.y*t+ca) * y + (u.y*u.z*t - u.x*sa) * z, (u.z*u.x*t - u.y*sa) * x + (u.y*u.z*t + u.x*sa) * y + (u.z*u.z*t + ca) * z);} //rotates by arbitrary vector arbitrary amount (http://www.cprogramming.com/tutorial/3d/rotation.html (Errors! LH one is actually RH one))
	inline Vec3D Rot(CQuat& Q) const; //below CQuat for linking sake...

	void RotZ(const vfloat a) {vfloat xt =  x*cos(a) - y*sin(a); vfloat yt = x*sin(a) + y*cos(a); x = xt; y = yt;} //rotates about Z axis "a" radians
	void RotY(const vfloat a) {vfloat xt =  x*cos(a) + z*sin(a); vfloat zt = -x*sin(a) + z*cos(a); x = xt; z = zt;} //rotates about Y axis "a" radians
	void RotX(const vfloat a) {vfloat yt =  y*cos(a) + z*sin(a); vfloat zt = -y*sin(a) + z*cos(a); y = yt; z = zt;} //rotates about X axis "a" radians

	//Vector operations (don't change this object!)
	inline Vec3D Cross(const Vec3D& v) const {return Vec3D(y*v.z-z*v.y,z*v.x-x*v.z,x*v.y-y*v.x);} //Cross product
	inline vfloat Dot(const Vec3D& v) const {return (x * v.x + y * v.y + z * v.z);} //Dot product
	inline Vec3D Abs() const {return Vec3D(fabs(x),fabs(y),fabs(z));} //Absolute value
	inline Vec3D Normalized() const {	vfloat l = sqrt(x*x+y*y+z*z); return (l>0?(*this)/l:(*this));} //returns normalized vec
	inline Vec3D ProjXY() const {	return Vec3D(x,y,0);} //projection into the xy plane
	inline vfloat Length() const {return sqrt(x*x+y*y+z*z);} //length of the vector
	inline vfloat Length2() const {return (x*x+y*y+z*z);} //length squared of the vector
	inline Vec3D Min(const Vec3D& s) const {return Vec3D(x<s.x ? x:s.x, y<s.y ? y:s.y, z<s.z ? z:s.z);} //min vector of the two
	inline Vec3D Max(const Vec3D& s) const {return Vec3D(x>s.x ? x:s.x, y>s.y ? y:s.y, z>s.z ? z:s.z);} //max vector of the two
	inline vfloat Min() const {vfloat Min1 = (x<y ? x:y); return (z<Min1 ? z:Min1);} //minimum element of this vector
	inline vfloat Max() const {vfloat Max1 = (x>y ? x:y); return (z>Max1 ? z:Max1);} //max element of this vector
	inline Vec3D Scale(const Vec3D& v) const {return Vec3D(x*v.x, y*v.y, z*v.z);} //scales by another vector
	inline Vec3D ScaleInv(const Vec3D& v) const {return Vec3D(x/v.x, y/v.y, z/v.z);} //scales by inverse of another vector
	inline vfloat Dist(const Vec3D& v) const {return sqrt(Dist2(v));} //distance from another vector
	inline vfloat Dist2(const Vec3D& v) const {return (v.x-x)*(v.x-x)+(v.y-y)*(v.y-y)+(v.z-z)*(v.z-z);} //distance from another vector
	inline vfloat AlignWith(const Vec3D target, Vec3D& rotax) const {Vec3D thisvec = Normalized(); Vec3D targvec = target.Normalized(); Vec3D rotaxis = thisvec.Cross(targvec); if (rotaxis.Length2() == 0) {rotaxis=target.ArbitraryNormal();} rotax = rotaxis.Normalized(); return acos(thisvec.Dot(targvec));} //returns vector (rotax) and amount (return val) to align this vector with target vector
	inline Vec3D ArbitraryNormal() const {Vec3D n = Normalized(); if (fabs(n.x) <= fabs(n.y) && fabs(n.x) <= fabs(n.z)){n.x = 1;} else if (fabs(n.y) <= fabs(n.x) && fabs(n.y) <= fabs(n.z)){n.y = 1;}	else {n.z = 1;}	return Cross(n).Normalized();} //generate arbitrary normal
};

//quaternion properties (for my reference)
//1) To rotate a vector V, form a quaternion with w = 0; To rotate by Quaternion Q, do Q*V*Q.Conjugate() and trim off the w component.
//2) To do multiple rotations: To Rotate by Q1 THEN Q2, Q2*Q1*V*Q1.Conjugate*Q2.Conjugate(), or make a Qtot = Q2*Q1 and do Qtot*V*Qtot.Conjucate()
//3) Q1*Q1.Conjugate - Identity
//4) To do a reverse rotation Q1, just do Q1.conjugate*V*Q1
//http://www.cprogramming.com/tutorial/3d/quaternions.html

class CQuat // : public Vec3D //extending Vec3D saves us reimplementing some stuff, I think? this is not a comprehensive quaternion class at this point...
{
public:
	CQuat(void) {Clear();}
	~CQuat(void){};
	
	CQuat(const vfloat dw, const vfloat dx, const vfloat dy, const vfloat dz) {w=dw; x=dx; y=dy; z=dz;} //constructor
	CQuat(const CQuat& Quat) {*this = Quat;} //copy constructor
	CQuat(const Vec3D& VecIn) {w = 0; x = VecIn.x; y = VecIn.y; z = VecIn.z;}
	CQuat(const vfloat angle, const Vec3D &axis){Clear(); const vfloat a = angle * (vfloat)0.5f; const vfloat s = sin(a); const vfloat c = cos(a); w = c; x = axis.x * s; y = axis.y * s; z = axis.z * s;};
	CQuat(const Vec3D &RotateFrom, const Vec3D &RotateTo){
		Vec3D Axis = RotateFrom.Cross(RotateTo);
		Axis.Normalize();
		vfloat DotP = RotateFrom.Dot(RotateTo)/(RotateFrom.Length()*RotateTo.Length());
		if (DotP > (vfloat) 0.999999999) {*this = CQuat(1,0,0,0); return;}
		if (DotP < (vfloat) -0.999999999) {*this = CQuat(0,1,0,0); return;}
		*this = CQuat(acos(DotP), Axis);
	};
	
	inline CQuat& operator=(const CQuat& s) {w = s.w; x = s.x; y = s.y; z = s.z; return *this; }; //overload equals
	inline void Clear() {w=1; x=0; y=0; z=0;} // for (int i=0; i<3; i++){for (int j=0; j<3; j++){M[i][j] = 0;iM[i][j] = 0;}}};

	inline Vec3D ToVec(void) const {return Vec3D(x, y, z);} //shouldnt be necessary... should be able to just set equal...

	friend CQuat operator+(const CQuat& s1, const CQuat& s2) {return CQuat(s1.w+s2.w, s1.x+s2.x, s1.y+s2.y, s1.z+s2.z);} //Plus
	CQuat& operator+=(const CQuat& s) {w += s.w; x += s.x; y += s.y; z += s.z; return *this;} //add and set
	friend CQuat operator-(const CQuat& s1, const CQuat& s2) {return CQuat(s1.w-s2.w, s1.x-s2.x, s1.y-s2.y, s1.z-s2.z);} //Minus
	CQuat& operator-=(const CQuat& s) {w -= s.w; x -= s.x; y -= s.y; z -= s.z; return *this;} //subract and set

	inline CQuat operator*(const vfloat f) {return CQuat(w*f, x*f, y*f, z*f);}; //scalar multiplication
	friend CQuat operator*(const vfloat f, const CQuat v) {return CQuat(v.w*f, v.x*f, v.y*f, v.z*f);};
	inline CQuat operator*(const CQuat& f) {return CQuat(w*f.w - x*f.x - y*f.y - z*f.z, w*f.x + x*f.w + y*f.z - z*f.y, w*f.y - x*f.z + y*f.w + z*f.x, w*f.z + x*f.y - y*f.x + z*f.w);}; //overload Quaternion multiplication!
	operator Vec3D() {return Vec3D(x, y, z);};

	friend bool operator==(const CQuat q1, const CQuat& q2) {return (q1.w==q2.w && q1.x==q2.x && q1.y==q2.y && q1.z==q2.z);} //Is equal


	inline vfloat Length() const {return sqrt(Length2());} //length of the vector
	inline vfloat Length2() const {return (w*w+x*x+y*y+z*z);} //length squared of the vector
	vfloat Normalize(void) {vfloat l = Length(); if (l == 0){w = 1; x = 0; y = 0; z = 0;} else if (l > 0) {w /= l; x /= l; y /= l; z /= l;} return l;};
	CQuat Inverse(void) const {vfloat n = w*w + x*x + y*y + z*z; return CQuat(w/n, -x/n, -y/n, -z/n); };
	CQuat Conjugate(void) const {return CQuat(w, -x, -y, -z);};

	vfloat w, x, y, z;

#ifdef MATRIX_OUTPUT_ENABLED
//	vfloat M [3][3];
//	vfloat iM [3][3];
//	void CalcMatrix(void) {M[0][0] = 1.0f-2.0f*(y*y+z*z); M[0][1] = 2.0f*(y*x-z*w); M[0][2] = 2.0f*(z*x+y*w); M[1][1] = 1.0f-2.0f*(x*x+z*z); M[1][2] = 2.0f*(z*y-x*w); M[2][2] = 1.0f-2.0f*(x*x+y*y); M[1][0] = M[0][1]; M[2][1] = M[1][2]; M[2][0] = M[0][2];};
//	void CalciMatrix(void) {vfloat determinant = -M[0][2]*M[1][1]*M[2][0] + M[0][1]*M[1][2]*M[2][0] + M[0][2]*M[1][0]*M[2][1] - M[0][0]*M[1][2]*M[2][1] - M[0][1]*M[1][0]*M[2][2] + M[0][0]*M[1][1]*M[2][2]; vfloat k = 1.0 / determinant; iM[0][0] = (M[1][1]*M[2][2] - M[2][1]*M[1][2])*k; iM[0][1] = (M[2][1]*M[0][2] - M[0][1]*M[2][2])*k; iM[0][2] = (M[0][1]*M[1][2] - M[1][1]*M[0][2])*k; iM[1][0] = (M[1][2]*M[2][0] - M[2][2]*M[1][0])*k; iM[1][1] = (M[2][2]*M[0][0] - M[0][2]*M[2][0])*k; iM[1][2] = (M[0][2]*M[1][0] - M[1][2]*M[0][0])*k; iM[2][0] = (M[1][0]*M[2][1] - M[2][0]*M[1][1])*k; iM[2][1] = (M[2][0]*M[0][1] - M[0][0]*M[2][1])*k; iM[2][2] = (M[0][0]*M[1][1] - M[1][0]*M[0][1])*k;};
#endif
	void AngleAxis(vfloat &angle, Vec3D &axis) const  {/*if (w>1.0) w = 1.0; */vfloat squareLength = x*x + y*y + z*z; if (squareLength>0.0000000001f){angle = (vfloat)2.0f * acos(w>1?1:w); vfloat inverseLength = pow(squareLength, (vfloat)-0.5f); axis.x = x * inverseLength; axis.y = y * inverseLength; axis.z = z * inverseLength;} else{angle = (vfloat)0.0f; axis.x = (vfloat)1.0f; axis.y = (vfloat)0.0f; axis.z = (vfloat)0.0f;}};

	void AngleAxis(vfloat *pAngle, vfloat* pAxisX, vfloat* pAxisY, vfloat* pAxisZ) const 
	{
	//	if (w>1.0) w = 1.0; 
		vfloat squareLength = x*x + y*y + z*z;
		if (squareLength>(vfloat)0.0000000001f){
			*pAngle = (vfloat)2.0f * (vfloat) acos(w>1?1:w);
			vfloat inverseLength = (vfloat) pow(squareLength, (vfloat)-0.5f);
			*pAxisX = x * inverseLength;
			*pAxisY = y * inverseLength;
			*pAxisZ = z * inverseLength;
		}
		else{
			*pAngle = 0.0f; *pAxisX = 1.0f; *pAxisY = 0.0f; *pAxisZ = 0.0f;}
	};

};


inline Vec3D Vec3D::Rot(CQuat& Q) const {return (Q*CQuat(*this)*Q.Conjugate()).ToVec();}


#endif //_VEC3D_H
